"""

Builtin events sent by Textual.

Events may be marked as "Bubbles" and "Verbose".
See the [events guide](/guide/events/#bubbling) for an explanation of bubbling.
Verbose events are excluded from the textual console, unless you explicitly request them with the `-v` switch as follows:

```
textual console -v
```
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Type, TypeVar

import rich.repr
from rich.style import Style
from typing_extensions import Self

from textual._types import CallbackType
from textual.geometry import Offset, Size
from textual.keys import _get_key_aliases
from textual.message import Message

MouseEventT = TypeVar("MouseEventT", bound="MouseEvent")

if TYPE_CHECKING:
    from textual.dom import DOMNode
    from textual.timer import Timer as TimerClass
    from textual.timer import TimerCallback
    from textual.widget import Widget


@rich.repr.auto
class Event(Message):
    """The base class for all events."""


@rich.repr.auto
class Callback(Event, bubble=False, verbose=True):
    """Sent by Textual to invoke a callback
    (see [call_next][textual.message_pump.MessagePump.call_next] and
    [call_later][textual.message_pump.MessagePump.call_later]).
    """

    def __init__(self, callback: CallbackType) -> None:
        self.callback = callback
        super().__init__()

    def __rich_repr__(self) -> rich.repr.Result:
        yield "callback", self.callback


@dataclass
class CursorPosition(Event, bubble=False):
    """Internal event used to retrieve the terminal's cursor position."""

    x: int
    y: int


class Load(Event, bubble=False):
    """
    Sent when the App is running but *before* the terminal is in application mode.

    Use this event to run any setup that doesn't require any visuals such as loading
    configuration and binding keys.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Idle(Event, bubble=False):
    """Sent when there are no more items in the message queue.

    This is a pseudo-event in that it is created by the Textual system and doesn't go
    through the usual message queue.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Action(Event):
    __slots__ = ["action"]

    def __init__(self, action: str) -> None:
        super().__init__()
        self.action = action

    def __rich_repr__(self) -> rich.repr.Result:
        yield "action", self.action


class Resize(Event, bubble=False):
    """Sent when the app or widget has been resized.

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        size: The new size of the Widget.
        virtual_size: The virtual size (scrollable size) of the Widget.
        container_size: The size of the Widget's container widget.
    """

    __slots__ = ["size", "virtual_size", "container_size"]

    def __init__(
        self,
        size: Size,
        virtual_size: Size,
        container_size: Size | None = None,
        pixel_size: Size | None = None,
    ) -> None:
        self.size = size
        """The new size of the Widget."""
        self.virtual_size = virtual_size
        """The virtual size (scrollable size) of the Widget."""
        self.container_size = size if container_size is None else container_size
        """The size of the Widget's container widget."""
        self.pixel_size = pixel_size
        """Size of terminal window in pixels if known, or `None` if not known."""
        super().__init__()

    @classmethod
    def from_dimensions(
        cls, cells: tuple[int, int], pixels: tuple[int, int] | None
    ) -> Resize:
        """Construct from basic dimensions.

        Args:
            cells: tuple of (<width>, <height>) in cells.
            pixels: tuple of (<width>, <height>) in pixels if known, or `None` if not known.

        """
        size = Size(*cells)
        pixel_size = Size(*pixels) if pixels is not None else None
        return Resize(size, size, size, pixel_size)

    def can_replace(self, message: "Message") -> bool:
        return isinstance(message, Resize)

    def __rich_repr__(self) -> rich.repr.Result:
        yield "size", self.size
        yield "virtual_size", self.virtual_size, self.size
        yield "container_size", self.container_size, self.size
        yield "pixel_size", self.pixel_size, None


class Compose(Event, bubble=False, verbose=True):
    """Sent to a widget to request it to compose and mount children.

    This event is used internally by Textual.
    You won't typically need to explicitly handle it,

    - [ ] Bubbles
    - [X] Verbose
    """


class Mount(Event, bubble=False, verbose=False):
    """Sent when a widget is *mounted* and may receive messages.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Unmount(Event, bubble=False, verbose=False):
    """Sent when a widget is unmounted and may no longer receive messages.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Show(Event, bubble=False):
    """Sent when a widget is first displayed.

    - [ ] Bubbles
    - [ ] Verbose
    """


class Hide(Event, bubble=False):
    """Sent when a widget has been hidden.

    - [ ] Bubbles
    - [ ] Verbose

    Sent when any of the following conditions apply:

    - The widget is removed from the DOM.
    - The widget is no longer displayed because it has been scrolled or clipped from the terminal or its container.
    - The widget has its `display` attribute set to `False`.
    - The widget's `display` style is set to `"none"`.
    """


class Ready(Event, bubble=False):
    """Sent to the `App` when the DOM is ready and the first frame has been displayed.

    - [ ] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class MouseCapture(Event, bubble=False):
    """Sent when the mouse has been captured.

    - [ ] Bubbles
    - [ ] Verbose

    When a mouse has been captured, all further mouse events will be sent to the capturing widget.

    Args:
        mouse_position: The position of the mouse when captured.
    """

    def __init__(self, mouse_position: Offset) -> None:
        super().__init__()
        self.mouse_position = mouse_position
        """The position of the mouse when captured."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield None, self.mouse_position


@rich.repr.auto
class MouseRelease(Event, bubble=False):
    """Mouse has been released.

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        mouse_position: The position of the mouse when released.
    """

    def __init__(self, mouse_position: Offset) -> None:
        super().__init__()
        self.mouse_position = mouse_position
        """The position of the mouse when released."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield None, self.mouse_position


class InputEvent(Event):
    """Base class for input events."""


@rich.repr.auto
class Key(InputEvent):
    """Sent when the user hits a key on the keyboard.

    - [X] Bubbles
    - [ ] Verbose

    Args:
        key: The key that was pressed.
        character: A printable character or `None` if it is not printable.
    """

    __slots__ = ["key", "character", "aliases"]

    def __init__(self, key: str, character: str | None) -> None:
        super().__init__()
        self.key = key
        """The key that was pressed."""
        self.character = (
            (key if len(key) == 1 else None) if character is None else character
        )
        """A printable character or ``None`` if it is not printable."""
        self.aliases: list[str] = _get_key_aliases(key)
        """The aliases for the key, including the key itself."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield "key", self.key
        yield "character", self.character
        yield "name", self.name
        yield "is_printable", self.is_printable
        yield "aliases", self.aliases, [self.key]

    @property
    def name(self) -> str:
        """Name of a key suitable for use as a Python identifier."""
        return _key_to_identifier(self.key).lower()

    @property
    def name_aliases(self) -> list[str]:
        """The corresponding name for every alias in `aliases` list."""
        return [_key_to_identifier(key) for key in self.aliases]

    @property
    def is_printable(self) -> bool:
        """Check if the key is printable (produces a unicode character).

        Returns:
            `True` if the key is printable.
        """
        return False if self.character is None else self.character.isprintable()


def _key_to_identifier(key: str) -> str:
    """Convert the key string to a name suitable for use as a Python identifier."""
    key_no_modifiers = key.split("+")[-1]
    if len(key_no_modifiers) == 1 and key_no_modifiers.isupper():
        if "+" in key:
            key = f"{key.rpartition('+')[0]}+upper_{key_no_modifiers}"
        else:
            key = f"upper_{key_no_modifiers}"
    return key.replace("+", "_").lower()


@rich.repr.auto
class MouseEvent(InputEvent, bubble=True):
    """Sent in response to a mouse event.

    - [X] Bubbles
    - [ ] Verbose

    Args:
        widget: The widget under the mouse.
        x: The relative x coordinate.
        y: The relative y coordinate.
        delta_x: Change in x since the last message.
        delta_y: Change in y since the last message.
        button: Indexed of the pressed button.
        shift: True if the shift key is pressed.
        meta: True if the meta key is pressed.
        ctrl: True if the ctrl key is pressed.
        screen_x: The absolute x coordinate.
        screen_y: The absolute y coordinate.
        style: The Rich Style under the mouse cursor.
    """

    __slots__ = [
        "widget",
        "_x",
        "_y",
        "_delta_x",
        "_delta_y",
        "button",
        "shift",
        "meta",
        "ctrl",
        "_screen_x",
        "_screen_y",
        "_style",
    ]

    def __init__(
        self,
        widget: Widget | None,
        x: float,
        y: float,
        delta_x: int,
        delta_y: int,
        button: int,
        shift: bool,
        meta: bool,
        ctrl: bool,
        screen_x: float | None = None,
        screen_y: float | None = None,
        style: Style | None = None,
    ) -> None:
        super().__init__()
        self.widget: Widget | None = widget
        """The widget under the mouse at the time of a click."""
        self._x = x
        """The relative x coordinate."""
        self._y = y
        """The relative y coordinate."""
        self._delta_x = delta_x
        """Change in x since the last message."""
        self._delta_y = delta_y
        """Change in y since the last message."""
        self.button = button
        """Indexed of the pressed button."""
        self.shift = shift
        """`True` if the shift key is pressed."""
        self.meta = meta
        """`True` if the meta key is pressed."""
        self.ctrl = ctrl
        """`True` if the ctrl key is pressed."""
        self._screen_x = x if screen_x is None else screen_x
        """The absolute x coordinate."""
        self._screen_y = y if screen_y is None else screen_y
        """The absolute y coordinate."""
        self._style = style or Style()

    @property
    def x(self) -> int:
        """The relative X coordinate of the cell under the mouse."""
        return int(self._x)

    @property
    def y(self) -> int:
        """The relative Y coordinate of the cell under the mouse."""
        return int(self._y)

    @property
    def delta_x(self) -> int:
        """Change in `x` since last message."""
        return self._delta_x

    @property
    def delta_y(self) -> int:
        """Change in `y` since the last message."""
        return self._delta_y

    @property
    def screen_x(self) -> int:
        """X coordinate of the cell relative to top left of screen."""
        return int(self._screen_x)

    @property
    def screen_y(self) -> int:
        """Y coordinate of the cell relative to top left of screen."""
        return int(self._screen_y)

    @property
    def pointer_x(self) -> float:
        """The relative X coordinate of the pointer."""
        return self._x

    @property
    def pointer_y(self) -> float:
        """The relative Y coordinate of the pointer."""
        return self._y

    @property
    def pointer_screen_x(self) -> float:
        """The X coordinate of the pointer relative to the screen."""
        return self._screen_x

    @property
    def pointer_screen_y(self) -> float:
        """The Y coordinate of the pointer relative to the screen."""
        return self._screen_y

    @classmethod
    def from_event(
        cls: Type[MouseEventT], widget: Widget, event: MouseEvent
    ) -> MouseEventT:
        new_event = cls(
            widget,
            event._x,
            event._y,
            event._delta_x,
            event._delta_y,
            event.button,
            event.shift,
            event.meta,
            event.ctrl,
            event._screen_x,
            event._screen_y,
            event._style,
        )
        return new_event

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.widget
        yield "x", self.x
        yield "y", self.y
        yield "pointer_x", self.pointer_x
        yield "pointer_y", self.pointer_y
        yield "delta_x", self.delta_x, 0
        yield "delta_y", self.delta_y, 0
        if self.screen_x != self.x:
            yield "screen_x", self._screen_x
        if self.screen_y != self.y:
            yield "screen_y", self._screen_y
        yield "button", self.button, 0
        yield "shift", self.shift, False
        yield "meta", self.meta, False
        yield "ctrl", self.ctrl, False
        if self.style:
            yield "style", self.style

    @property
    def control(self) -> Widget | None:
        return self.widget

    @property
    def offset(self) -> Offset:
        """The mouse coordinate as an offset.

        Returns:
            Mouse coordinate.
        """
        return Offset(self.x, self.y)

    @property
    def screen_offset(self) -> Offset:
        """Mouse coordinate relative to the screen."""
        return Offset(self.screen_x, self.screen_y)

    @property
    def delta(self) -> Offset:
        """Mouse coordinate delta (change since last event)."""
        return Offset(self.delta_x, self.delta_y)

    @property
    def style(self) -> Style:
        """The (Rich) Style under the cursor."""
        return self._style or Style()

    @style.setter
    def style(self, style: Style) -> None:
        self._style = style

    def get_content_offset(self, widget: Widget) -> Offset | None:
        """Get offset within a widget's content area, or None if offset is not in content (i.e. padding or border).

        Args:
            widget: Widget receiving the event.

        Returns:
            An offset where the origin is at the top left of the content area.
        """
        if self.screen_offset not in widget.content_region:
            return None
        return self.get_content_offset_capture(widget)

    def get_content_offset_capture(self, widget: Widget) -> Offset:
        """Get offset from a widget's content area.

        This method works even if the offset is outside the widget content region.

        Args:
            widget: Widget receiving the event.

        Returns:
            An offset where the origin is at the top left of the content area.
        """
        return self.offset - widget.gutter.top_left

    def _apply_offset(self, x: int, y: int) -> MouseEvent:
        return self.__class__(
            self.widget,
            x=self._x + x,
            y=self._y + y,
            delta_x=self._delta_x,
            delta_y=self._delta_y,
            button=self.button,
            shift=self.shift,
            meta=self.meta,
            ctrl=self.ctrl,
            screen_x=self._screen_x,
            screen_y=self._screen_y,
            style=self.style,
        )


@rich.repr.auto
class MouseMove(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse cursor moves.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseDown(MouseEvent, bubble=True, verbose=True):
    """Sent when a mouse button is pressed.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseUp(MouseEvent, bubble=True, verbose=True):
    """Sent when a mouse button is released.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseScrollDown(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse wheel is scrolled *down*.

    - [X] Bubbles
    - [X] Verbose
    """


@rich.repr.auto
class MouseScrollUp(MouseEvent, bubble=True, verbose=True):
    """Sent when the mouse wheel is scrolled *up*.

    - [X] Bubbles
    - [X] Verbose
    """


class Click(MouseEvent, bubble=True):
    """Sent when a widget is clicked.

    - [X] Bubbles
    - [ ] Verbose

    Args:
        chain: The number of clicks in the chain. 2 is a double click, 3 is a triple click, etc.
    """

    def __init__(
        self,
        widget: Widget | None,
        x: int,
        y: int,
        delta_x: int,
        delta_y: int,
        button: int,
        shift: bool,
        meta: bool,
        ctrl: bool,
        screen_x: int | None = None,
        screen_y: int | None = None,
        style: Style | None = None,
        chain: int = 1,
    ) -> None:
        super().__init__(
            widget,
            x,
            y,
            delta_x,
            delta_y,
            button,
            shift,
            meta,
            ctrl,
            screen_x,
            screen_y,
            style,
        )
        self.chain = chain

    @classmethod
    def from_event(
        cls: Type[Self],
        widget: Widget,
        event: MouseEvent,
        chain: int = 1,
    ) -> Self:
        new_event = cls(
            widget,
            event.x,
            event.y,
            event.delta_x,
            event.delta_y,
            event.button,
            event.shift,
            event.meta,
            event.ctrl,
            event.screen_x,
            event.screen_y,
            event._style,
            chain=chain,
        )
        return new_event

    def _apply_offset(self, x: int, y: int) -> Self:
        return self.__class__(
            self.widget,
            x=self.x + x,
            y=self.y + y,
            delta_x=self.delta_x,
            delta_y=self.delta_y,
            button=self.button,
            shift=self.shift,
            meta=self.meta,
            ctrl=self.ctrl,
            screen_x=self.screen_x,
            screen_y=self.screen_y,
            style=self.style,
            chain=self.chain,
        )

    def __rich_repr__(self) -> rich.repr.Result:
        yield from super().__rich_repr__()
        yield "chain", self.chain


@rich.repr.auto
class Timer(Event, bubble=False, verbose=True):
    """Sent in response to a timer.

    - [ ] Bubbles
    - [X] Verbose
    """

    __slots__ = ["timer", "time", "count", "callback"]

    def __init__(
        self,
        timer: "TimerClass",
        time: float,
        count: int = 0,
        callback: TimerCallback | None = None,
    ) -> None:
        super().__init__()
        self.timer = timer
        self.time = time
        self.count = count
        self.callback = callback

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.timer.name
        yield "count", self.count


class Enter(Event, bubble=True, verbose=True):
    """Sent when the mouse is moved over a widget.

    Note that this event bubbles, so a widget may receive this event when the mouse
    moves over a child widget. Check the `node` attribute for the widget directly under
    the mouse.

    - [X] Bubbles
    - [X] Verbose
    """

    __slots__ = ["node"]

    def __init__(self, node: DOMNode) -> None:
        self.node = node
        """The node directly under the mouse."""
        super().__init__()

    @property
    def control(self) -> DOMNode:
        """Alias for the `node` under the mouse."""
        return self.node


class Leave(Event, bubble=True, verbose=True):
    """Sent when the mouse is moved away from a widget, or if a widget is
    programmatically disabled while hovered.

    Note that this widget bubbles, so a widget may receive Leave events for any child widgets.
    Check the `node` parameter for the original widget that was previously under the mouse.


    - [X] Bubbles
    - [X] Verbose
    """

    __slots__ = ["node"]

    def __init__(self, node: DOMNode) -> None:
        self.node = node
        """The node that was previously directly under the mouse."""
        super().__init__()

    @property
    def control(self) -> DOMNode:
        """Alias for the `node` that was previously under the mouse."""
        return self.node


class Focus(Event, bubble=False):
    """Sent when a widget is focussed.

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        from_app_focus: True if this focus event has been sent because the app itself has
            regained focus (via an AppFocus event). False if the focus came from within
            the Textual app (e.g. via the user pressing tab or a programmatic setting
            of the focused widget).
    """

    def __init__(self, from_app_focus: bool = False) -> None:
        self.from_app_focus = from_app_focus
        super().__init__()

    def __rich_repr__(self) -> rich.repr.Result:
        yield from super().__rich_repr__()
        yield "from_app_focus", self.from_app_focus


class Blur(Event, bubble=False):
    """Sent when a widget is blurred (un-focussed).

    - [ ] Bubbles
    - [ ] Verbose
    """


class AppFocus(Event, bubble=False):
    """Sent when the app has focus.

    - [ ] Bubbles
    - [ ] Verbose

    Note:
        Only available when running within a terminal that supports
        `FocusIn`, or when running via textual-web.
    """


class AppBlur(Event, bubble=False):
    """Sent when the app loses focus.

    - [ ] Bubbles
    - [ ] Verbose

    Note:
        Only available when running within a terminal that supports
        `FocusOut`, or when running via textual-web.
    """


@dataclass
class DescendantFocus(Event, bubble=True, verbose=True):
    """Sent when a child widget is focussed.

    - [X] Bubbles
    - [X] Verbose
    """

    widget: Widget
    """The widget that was focused."""

    @property
    def control(self) -> Widget:
        """The widget that was focused (alias of `widget`)."""
        return self.widget


@dataclass
class DescendantBlur(Event, bubble=True, verbose=True):
    """Sent when a child widget is blurred.

    - [X] Bubbles
    - [X] Verbose
    """

    widget: Widget
    """The widget that was blurred."""

    @property
    def control(self) -> Widget:
        """The widget that was blurred (alias of `widget`)."""
        return self.widget


@rich.repr.auto
class Paste(Event, bubble=True):
    """Event containing text that was pasted into the Textual application.
    This event will only appear when running in a terminal emulator that supports
    bracketed paste mode. Textual will enable bracketed pastes when an app starts,
    and disable it when the app shuts down.

    - [X] Bubbles
    - [ ] Verbose


    Args:
        text: The text that has been pasted.
    """

    def __init__(self, text: str) -> None:
        super().__init__()
        self.text = text
        """The text that was pasted."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield "text", self.text


class ScreenResume(Event, bubble=False):
    """Sent to screen that has been made active.

    - [ ] Bubbles
    - [ ] Verbose
    """


class ScreenSuspend(Event, bubble=False):
    """Sent to screen when it is no longer active.

    - [ ] Bubbles
    - [ ] Verbose
    """


@rich.repr.auto
class Print(Event, bubble=False):
    """Sent to a widget that is capturing [`print`][print].

    - [ ] Bubbles
    - [ ] Verbose

    Args:
        text: Text that was printed.
        stderr: `True` if the print was to stderr, or `False` for stdout.

    Note:
        Python's [`print`][print] output can be captured with
        [`App.begin_capture_print`][textual.app.App.begin_capture_print].
    """

    def __init__(self, text: str, stderr: bool = False) -> None:
        super().__init__()
        self.text = text
        """The text that was printed."""
        self.stderr = stderr
        """`True` if the print was to stderr, or `False` for stdout."""

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.text
        yield self.stderr


@dataclass
class DeliveryComplete(Event, bubble=False):
    """Sent to App when a file has been delivered."""

    key: str
    """The delivery key associated with the delivery.
    
    This is the same key that was returned by `App.deliver_text`/`App.deliver_binary`.
    """

    path: Path | None = None
    """The path where the file was saved, or `None` if the path is not available, for
    example if the file was delivered via web browser.
    """

    name: str | None = None
    """Optional name returned to the app to identify the download."""


@dataclass
class DeliveryFailed(Event, bubble=False):
    """Sent to App when a file delivery fails."""

    key: str
    """The delivery key associated with the delivery."""

    exception: BaseException
    """The exception that was raised during the delivery."""

    name: str | None = None
    """Optional name returned to the app to identify the download."""



import json

COUNTRIES = [
    "Afghanistan",
    "Albania",
    "Algeria",
    "Andorra",
    "Angola",
    "Antigua and Barbuda",
    "Argentina",
    "Armenia",
    "Australia",
    "Austria",
    "Azerbaijan",
    "Bahamas",
    "Bahrain",
    "Bangladesh",
    "Barbados",
    "Belarus",
    "Belgium",
    "Belize",
    "Benin",
    "Bhutan",
    "Bolivia",
    "Bosnia and Herzegovina",
    "Botswana",
    "Brazil",
    "Brunei",
    "Bulgaria",
    "Burkina Faso",
    "Burundi",
    "Cabo Verde",
    "Cambodia",
    "Cameroon",
    "Canada",
    "Central African Republic",
    "Chad",
    "Chile",
    "China",
    "Colombia",
    "Comoros",
    "Congo",
    "Costa Rica",
    "Croatia",
    "Cuba",
    "Cyprus",
    "Czech Republic",
    "Democratic Republic of the Congo",
    "Denmark",
    "Djibouti",
    "Dominica",
    "Dominican Republic",
    "East Timor",
    "Ecuador",
    "Egypt",
    "El Salvador",
    "Equatorial Guinea",
    "Eritrea",
    "Estonia",
    "Eswatini",
    "Ethiopia",
    "Fiji",
    "Finland",
    "France",
    "Gabon",
    "Gambia",
    "Georgia",
    "Germany",
    "Ghana",
    "Greece",
    "Grenada",
    "Guatemala",
    "Guinea",
    "Guinea-Bissau",
    "Guyana",
    "Haiti",
    "Honduras",
    "Hungary",
    "Iceland",
    "India",
    "Indonesia",
    "Iran",
    "Iraq",
    "Ireland",
    "Israel",
    "Italy",
    "Ivory Coast",
    "Jamaica",
    "Japan",
    "Jordan",
    "Kazakhstan",
    "Kenya",
    "Kiribati",
    "Kuwait",
    "Kyrgyzstan",
    "Laos",
    "Latvia",
    "Lebanon",
    "Lesotho",
    "Liberia",
    "Libya",
    "Liechtenstein",
    "Lithuania",
    "Luxembourg",
    "Madagascar",
    "Malawi",
    "Malaysia",
    "Maldives",
    "Mali",
    "Malta",
    "Marshall Islands",
    "Mauritania",
    "Mauritius",
    "Mexico",
    "Micronesia",
    "Moldova",
    "Monaco",
    "Mongolia",
    "Montenegro",
    "Morocco",
    "Mozambique",
    "Myanmar",
    "Namibia",
    "Nauru",
    "Nepal",
    "Netherlands",
    "New Zealand",
    "Nicaragua",
    "Niger",
    "Nigeria",
    "North Korea",
    "North Macedonia",
    "Norway",
    "Oman",
    "Pakistan",
    "Palau",
    "Palestine",
    "Panama",
    "Papua New Guinea",
    "Paraguay",
    "Peru",
    "Philippines",
    "Poland",
    "Portugal",
    "Qatar",
    "Romania",
    "Russia",
    "Rwanda",
    "Saint Kitts and Nevis",
    "Saint Lucia",
    "Saint Vincent and the Grenadines",
    "Samoa",
    "San Marino",
    "Sao Tome and Principe",
    "Saudi Arabia",
    "Senegal",
    "Serbia",
    "Seychelles",
    "Sierra Leone",
    "Singapore",
    "Slovakia",
    "Slovenia",
    "Solomon Islands",
    "Somalia",
    "South Africa",
    "South Korea",
    "South Sudan",
    "Spain",
    "Sri Lanka",
    "Sudan",
    "Suriname",
    "Sweden",
    "Switzerland",
    "Syria",
    "Taiwan",
    "Tajikistan",
    "Tanzania",
    "Thailand",
    "Togo",
    "Tonga",
    "Trinidad and Tobago",
    "Tunisia",
    "Turkey",
    "Turkmenistan",
    "Tuvalu",
    "Uganda",
    "Ukraine",
    "United Arab Emirates",
    "United Kingdom",
    "United States",
    "Uruguay",
    "Uzbekistan",
    "Vanuatu",
    "Vatican City",
    "Venezuela",
    "Vietnam",
    "Yemen",
    "Zambia",
    "Zimbabwe",
]
# Sort by length for auto-complete
COUNTRIES.sort(key=str.__len__)

# Thanks, Claude
MOVIES = """\
Date,Title,Genre,Director,Box Office (millions),Rating,Runtime (min)
1980-01-18,The Fog,Horror,John Carpenter,21,R,89
1980-02-15,Coal Miner's Daughter,Biography,Michael Apted,67,PG,124
1980-03-07,Little Miss Marker,Comedy,Walter Bernstein,12,PG,103
1980-04-11,The Long Riders,Western,Walter Hill,15,R,100
1980-05-21,The Empire Strikes Back,Sci-Fi,Irvin Kershner,538,PG,124
1980-06-13,The Blues Brothers,Comedy,John Landis,115,R,133
1980-07-02,Airplane!,Comedy,Jim Abrahams,83,PG,88
1980-08-01,Caddyshack,Comedy,Harold Ramis,39,R,98
1980-09-19,The Big Red One,War,Samuel Fuller,24,PG,113
1980-10-10,Private Benjamin,Comedy,Howard Zieff,69,R,109
1980-11-07,The Stunt Man,Action,Richard Rush,7,R,131
1980-12-19,Nine to Five,Comedy,Colin Higgins,103,PG,109
1981-01-23,Scanners,Horror,David Cronenberg,14,R,103
1981-02-20,The Final Conflict,Horror,Graham Baker,20,R,108
1981-03-20,Raiders of the Lost Ark,Action,Steven Spielberg,389,PG,115
1981-04-10,Excalibur,Fantasy,John Boorman,35,R,140
1981-05-22,Outland,Sci-Fi,Peter Hyams,17,R,109
1981-06-19,Superman II,Action,Richard Lester,108,PG,127
1981-07-17,Escape from New York,Sci-Fi,John Carpenter,25,R,99
1981-08-07,An American Werewolf in London,Horror,John Landis,30,R,97
1981-09-25,Continental Divide,Romance,Michael Apted,15,PG,103
1981-10-16,True Confessions,Drama,Ulu Grosbard,12,R,108
1981-11-20,Time Bandits,Fantasy,Terry Gilliam,42,PG,116
1981-12-04,Rollover,Drama,Alan J. Pakula,11,R,116
1982-01-15,The Beast Within,Horror,Philippe Mora,7,R,98
1982-02-12,Quest for Fire,Adventure,Jean-Jacques Annaud,20,R,100
1982-03-19,Porky's,Comedy,Bob Clark,105,R,94
1982-04-16,The Sword and the Sorcerer,Fantasy,Albert Pyun,39,R,99
1982-05-14,Conan the Barbarian,Fantasy,John Milius,68,R,129
1982-06-04,Star Trek II: The Wrath of Khan,Sci-Fi,Nicholas Meyer,97,PG,113
1982-06-11,E.T. the Extra-Terrestrial,Sci-Fi,Steven Spielberg,792,PG,115
1982-06-25,Blade Runner,Sci-Fi,Ridley Scott,33,R,117
1982-07-16,The World According to Garp,Comedy-Drama,George Roy Hill,29,R,136
1982-08-13,Fast Times at Ridgemont High,Comedy,Amy Heckerling,27,R,90
1982-09-17,The Challenge,Action,John Frankenheimer,9,R,108
1982-10-22,First Blood,Action,Ted Kotcheff,47,R,93
1982-11-12,The Man from Snowy River,Western,George Miller,20,PG,102
1982-12-08,48 Hrs.,Action,Walter Hill,79,R,96
1983-01-21,The Entity,Horror,Sidney J. Furie,13,R,125
1983-02-18,The Year of Living Dangerously,Drama,Peter Weir,10,PG,115
1983-03-25,The Outsiders,Drama,Francis Ford Coppola,25,PG,91
1983-04-22,Something Wicked This Way Comes,Horror,Jack Clayton,5,PG,95
1983-05-25,Return of the Jedi,Sci-Fi,Richard Marquand,475,PG,131
1983-06-17,Superman III,Action,Richard Lester,60,PG,125
1983-07-15,Class,Comedy,Lewis John Carlino,21,R,98
1983-08-19,Curse of the Pink Panther,Comedy,Blake Edwards,9,PG,109
1983-09-23,The Big Chill,Drama,Lawrence Kasdan,56,R,105
1983-10-07,The Right Stuff,Drama,Philip Kaufman,21,PG,193
1983-11-04,Deal of the Century,Comedy,William Friedkin,10,PG,99
1983-12-09,Scarface,Crime,Brian De Palma,65,R,170
1984-01-13,Terms of Endearment,Drama,James L. Brooks,108,PG,132
1984-02-17,Unfaithfully Yours,Comedy,Howard Zieff,12,PG,96
1984-03-16,Splash,Romance,Ron Howard,69,PG,111
1984-04-13,Friday the 13th: The Final Chapter,Horror,Joseph Zito,32,R,91
1984-05-04,Sixteen Candles,Comedy,John Hughes,23,PG,93
1984-06-08,Ghostbusters,Comedy,Ivan Reitman,295,PG,105
1984-07-06,The Last Starfighter,Sci-Fi,Nick Castle,28,PG,101
1984-08-10,Red Dawn,Action,John Milius,38,PG-13,114
1984-09-14,All of Me,Comedy,Carl Reiner,40,PG,93
1984-10-26,The Terminator,Sci-Fi,James Cameron,78,R,107
1984-11-16,Missing in Action,Action,Joseph Zito,22,R,101
1984-12-14,Dune,Sci-Fi,David Lynch,30,PG-13,137
1985-01-18,A Nightmare on Elm Street,Horror,Wes Craven,25,R,91
1985-02-15,The Breakfast Club,Drama,John Hughes,45,R,97
1985-03-29,Mask,Drama,Peter Bogdanovich,42,PG-13,120
1985-04-26,Code of Silence,Action,Andrew Davis,20,R,101
1985-05-22,Rambo: First Blood Part II,Action,George P. Cosmatos,150,R,96
1985-06-07,The Goonies,Adventure,Richard Donner,61,PG,114
1985-07-03,Back to the Future,Sci-Fi,Robert Zemeckis,381,PG,116
1985-08-16,Year of the Dragon,Crime,Michael Cimino,18,R,134
1985-09-20,Invasion U.S.A.,Action,Joseph Zito,17,R,107
1985-10-18,Silver Bullet,Horror,Daniel Attias,12,R,95
1985-11-22,Rocky IV,Drama,Sylvester Stallone,127,PG,91
1985-12-20,The Color Purple,Drama,Steven Spielberg,142,PG-13,154
1986-01-17,Iron Eagle,Action,Sidney J. Furie,24,PG-13,117
1986-02-21,Crossroads,Drama,Walter Hill,5,R,99
1986-03-21,Highlander,Fantasy,Russell Mulcahy,12,R,116
1986-04-18,Legend,Fantasy,Ridley Scott,15,PG,89
1986-05-16,Top Gun,Action,Tony Scott,357,PG,110
1986-06-27,Running Scared,Action,Peter Hyams,38,R,107
1986-07-18,Aliens,Sci-Fi,James Cameron,131,R,137
1986-08-08,Stand By Me,Drama,Rob Reiner,52,R,89
1986-09-19,Blue Velvet,Mystery,David Lynch,8,R,120
1986-10-24,The Name of the Rose,Mystery,Jean-Jacques Annaud,7,R,130
1986-11-21,An American Tail,Animation,Don Bluth,47,G,80
1986-12-19,Star Trek IV: The Voyage Home,Sci-Fi,Leonard Nimoy,109,PG,119
1987-01-23,Critical Condition,Comedy,Michael Apted,19,R,98
1987-02-20,Death Before Dishonor,Action,Terry Leonard,3,R,91
1987-03-13,Lethal Weapon,Action,Richard Donner,65,R,110
1987-04-10,Project X,Drama,Jonathan Kaplan,28,PG,108
1987-05-22,Beverly Hills Cop II,Action,Tony Scott,276,R,100
1987-06-19,Predator,Sci-Fi,John McTiernan,98,R,107
1987-07-17,RoboCop,Action,Paul Verhoeven,53,R,102
1987-08-14,No Way Out,Thriller,Roger Donaldson,35,R,114
1987-09-18,Fatal Beauty,Action,Tom Holland,12,R,104
1987-10-23,Fatal Attraction,Thriller,Adrian Lyne,320,R,119
1987-11-13,Running Man,Sci-Fi,Paul Michael Glaser,38,R,101
1987-12-18,Wall Street,Drama,Oliver Stone,43,R,126
1988-01-15,Return of the Living Dead Part II,Horror,Ken Wiederhorn,9,R,89
1988-02-12,Action Jackson,Action,Craig R. Baxley,20,R,96
1988-03-18,D.O.A.,Thriller,Rocky Morton,12,R,96
1988-04-29,Colors,Crime,Dennis Hopper,46,R,120
1988-05-20,Willow,Fantasy,Ron Howard,57,PG,126
1988-06-21,Big,Comedy,Penny Marshall,151,PG,104
1988-07-15,Die Hard,Action,John McTiernan,140,R,132
1988-08-05,Young Guns,Western,Christopher Cain,45,R,107
1988-09-16,Moon Over Parador,Comedy,Paul Mazursky,11,PG-13,103
1988-10-21,Halloween 4,Horror,Dwight H. Little,17,R,88
1988-11-11,Child's Play,Horror,Tom Holland,33,R,87
1988-12-21,Rain Man,Drama,Barry Levinson,172,R,133
1989-01-13,Deep Star Six,Sci-Fi,Sean S. Cunningham,8,R,99
1989-02-17,Bill & Ted's Excellent Adventure,Comedy,Stephen Herek,40,PG,90
1989-03-24,Leviathan,Sci-Fi,George P. Cosmatos,15,R,98
1989-04-14,Major League,Comedy,David S. Ward,49,R,107
1989-05-24,Indiana Jones and the Last Crusade,Action,Steven Spielberg,474,PG-13,127
1989-06-23,Batman,Action,Tim Burton,411,PG-13,126
1989-07-07,Lethal Weapon 2,Action,Richard Donner,227,R,114
1989-08-11,A Nightmare on Elm Street 5,Horror,Stephen Hopkins,22,R,89
1989-09-22,Black Rain,Action,Ridley Scott,46,R,125
1989-10-20,Look Who's Talking,Comedy,Amy Heckerling,140,PG-13,93
1989-11-17,All Dogs Go to Heaven,Animation,Don Bluth,27,G,84
1989-12-20,Tango & Cash,Action,Andrei Konchalovsky,63,R,104
"""

MOVIES_JSON = """{
  "decades": {
    "1980s": {
      "genres": {
        "action": {
          "franchises": {
            "terminator": {
              "name": "The Terminator",
              "movies": [
                {
                  "title": "The Terminator",
                  "year": 1984,
                  "director": "James Cameron",
                  "stars": ["Arnold Schwarzenegger", "Linda Hamilton", "Michael Biehn"],
                  "boxOffice": 78371200,
                  "quotes": ["I'll be back", "Come with me if you want to live"]
                }
              ]
            },
            "rambo": {
              "name": "Rambo",
              "movies": [
                {
                  "title": "First Blood",
                  "year": 1982,
                  "director": "Ted Kotcheff",
                  "stars": ["Sylvester Stallone", "Richard Crenna", "Brian Dennehy"],
                  "boxOffice": 47212904
                },
                {
                  "title": "Rambo: First Blood Part II",
                  "year": 1985,
                  "director": "George P. Cosmatos",
                  "stars": ["Sylvester Stallone", "Richard Crenna", "Charles Napier"],
                  "boxOffice": 150415432
                }
              ]
            }
          },
          "standalone_classics": {
            "die_hard": {
              "title": "Die Hard",
              "year": 1988,
              "director": "John McTiernan",
              "stars": ["Bruce Willis", "Alan Rickman", "Reginald VelJohnson"],
              "boxOffice": 140700000,
              "location": "Nakatomi Plaza",
              "quotes": ["Yippee-ki-yay, motherf***er"]
            },
            "predator": {
              "title": "Predator",
              "year": 1987,
              "director": "John McTiernan",
              "stars": ["Arnold Schwarzenegger", "Carl Weathers", "Jesse Ventura"],
              "boxOffice": 98267558,
              "location": "Val Verde jungle",
              "quotes": ["Get to the chopper!"]
            }
          },
          "common_themes": [
            "Cold War politics",
            "One man army",
            "Revenge plots",
            "Military operations",
            "Law enforcement"
          ],
          "typical_elements": {
            "weapons": ["M60 machine gun", "Desert Eagle", "Explosive arrows"],
            "vehicles": ["Military helicopters", "Muscle cars", "Tanks"],
            "locations": ["Urban jungle", "Actual jungle", "Industrial facilities"]
          }
        }
      }
    }
  },
  "metadata": {
    "total_movies": 4,
    "date_compiled": "2024",
    "box_office_total": 467654094,
    "most_frequent_actor": "Arnold Schwarzenegger",
    "most_frequent_director": "John McTiernan"
  }
}"""

MOVIES_TREE = json.loads(MOVIES_JSON)

DUNE_BIOS = [
    {
        "name": "Paul Atreides",
        "description": "Heir to House Atreides who becomes the Fremen messiah Muad'Dib. Born with extraordinary mental abilities due to Bene Gesserit breeding program.",
    },
    {
        "name": "Lady Jessica",
        "description": "Bene Gesserit concubine to Duke Leto and mother of Paul. Defied her order by bearing a son instead of a daughter, disrupting centuries of careful breeding.",
    },
    {
        "name": "Baron Vladimir Harkonnen",
        "description": "Cruel and corpulent leader of House Harkonnen, sworn enemy of House Atreides. Known for his cunning and brutality in pursuing power.",
    },
    {
        "name": "Leto Atreides",
        "description": "Noble Duke and father of Paul, known for his honor and just rule. Accepts governorship of Arrakis despite knowing it's likely a trap.",
    },
    {
        "name": "Stilgar",
        "description": "Leader of the Fremen Sietch Tabr, becomes a loyal supporter of Paul. Skilled warrior who helps train Paul in Fremen ways.",
    },
    {
        "name": "Chani",
        "description": "Fremen warrior and daughter of planetologist Liet-Kynes. Becomes Paul's concubine and true love after appearing in his prescient visions.",
    },
    {
        "name": "Thufir Hawat",
        "description": "Mentat and Master of Assassins for House Atreides. Serves three generations of Atreides with his superhuman computational skills.",
    },
    {
        "name": "Duncan Idaho",
        "description": "Swordmaster of the Ginaz, loyal to House Atreides. Known for his exceptional fighting skills and sacrifice to save Paul and Jessica.",
    },
    {
        "name": "Gurney Halleck",
        "description": "Warrior-troubadour of House Atreides, skilled with sword and baliset. Serves as Paul's weapons teacher and loyal friend.",
    },
    {
        "name": "Dr. Yueh",
        "description": "Suk doctor conditioned against taking human life, but betrays House Atreides after the Harkonnens torture his wife. Imperial Conditioning broken.",
    },
]



import pytest

from textual._dispatch_key import dispatch_key
from textual.app import App, ComposeResult
from textual.errors import DuplicateKeyHandlers
from textual.events import Key
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Button, Input, Label


class ValidWidget(Widget):
    called_by = None

    def key_x(self):
        self.called_by = self.key_x

    def key_ctrl_i(self):
        self.called_by = self.key_ctrl_i


async def test_dispatch_key_valid_key():
    widget = ValidWidget()
    result = await dispatch_key(widget, Key(key="x", character="x"))
    assert result is True
    assert widget.called_by == widget.key_x


async def test_dispatch_key_valid_key_alias():
    """When you press tab or ctrl+i, it comes through as a tab key event, but handlers for
    tab and ctrl+i are both considered valid."""
    widget = ValidWidget()
    result = await dispatch_key(widget, Key(key="tab", character="\t"))
    assert result is True
    assert widget.called_by == widget.key_ctrl_i


class DuplicateHandlersWidget(Widget):
    called_by = None

    def key_x(self):
        self.called_by = self.key_x

    def _key_x(self):
        self.called_by = self._key_x

    def key_tab(self):
        self.called_by = self.key_tab

    def key_ctrl_i(self):
        self.called_by = self.key_ctrl_i


async def test_dispatch_key_raises_when_conflicting_handler_aliases():
    """If you've got a handler for e.g. ctrl+i and a handler for tab, that's probably a mistake.
    In the terminal, they're the same thing, so we fail fast via exception here."""
    widget = DuplicateHandlersWidget()
    with pytest.raises(DuplicateKeyHandlers):
        await dispatch_key(widget, Key(key="tab", character="\t"))
    assert widget.called_by == widget.key_tab


class PreventTestApp(App):
    def __init__(self) -> None:
        self.input_changed_events = []
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Input()

    def on_input_changed(self, event: Input.Changed) -> None:
        self.input_changed_events.append(event)


async def test_message_queue_size():
    """Test message queue size property."""
    app = App()
    assert app.message_queue_size == 0

    class TestMessage(Message):
        pass

    async with app.run_test() as pilot:
        assert app.message_queue_size == 0
        app.post_message(TestMessage())
        assert app.message_queue_size == 1
        app.post_message(TestMessage())
        assert app.message_queue_size == 2
        # A pause will process all the messages
        await pilot.pause()
        assert app.message_queue_size == 0


async def test_prevent() -> None:
    app = PreventTestApp()

    async with app.run_test() as pilot:
        assert not app.input_changed_events
        input = app.query_one(Input)
        input.value = "foo"
        await pilot.pause()
        assert len(app.input_changed_events) == 1
        assert app.input_changed_events[0].value == "foo"

        with input.prevent(Input.Changed):
            input.value = "bar"

        await pilot.pause()
        assert len(app.input_changed_events) == 1
        assert app.input_changed_events[0].value == "foo"


async def test_prevent_with_call_next() -> None:
    """Test for https://github.com/Textualize/textual/issues/3166.

    Does a callback scheduled with `call_next` respect messages that
    were prevented when it was scheduled?
    """

    hits = 0

    class PreventTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Input()

        def change_input(self) -> None:
            self.query_one(Input).value += "a"

        def on_input_changed(self) -> None:
            nonlocal hits
            hits += 1

    app = PreventTestApp()
    async with app.run_test() as pilot:
        app.call_next(app.change_input)
        await pilot.pause()
        assert hits == 1

        with app.prevent(Input.Changed):
            app.call_next(app.change_input)
        await pilot.pause()
        assert hits == 1

        app.call_next(app.change_input)
        await pilot.pause()
        assert hits == 2


async def test_prevent_default():
    """Test that prevent_default doesn't apply when a message is bubbled."""

    app_button_pressed = False

    class MyButton(Button):
        def _on_button_pressed(self, event: Button.Pressed) -> None:
            event.prevent_default()

    class PreventApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MyButton("Press me")
            yield Label("No pressure")

        def on_button_pressed(self, event: Button.Pressed) -> None:
            nonlocal app_button_pressed
            app_button_pressed = True
            self.query_one(Label).update("Ouch!")

    app = PreventApp()
    async with app.run_test() as pilot:
        await pilot.click(MyButton)
        assert app_button_pressed



from typing import Callable

from textual.expand_tabs import get_tab_widths

__all__ = ["cell_len", "cell_width_to_column_index"]


cell_len: Callable[[str], int]
try:
    from rich.cells import cached_cell_len as cell_len
except ImportError:
    from rich.cells import cell_len


def cell_width_to_column_index(line: str, cell_width: int, tab_width: int) -> int:
    """Retrieve the column index corresponding to the given cell width.

    Args:
        line: The line of text to search within.
        cell_width: The cell width to convert to column index.
        tab_width: The tab stop width to expand tabs contained within the line.

    Returns:
        The column corresponding to the cell width.
    """
    column_index = 0
    total_cell_offset = 0
    for part, expanded_tab_width in get_tab_widths(line, tab_width):
        # Check if the click landed on a character within this part.
        for character in part:
            total_cell_offset += cell_len(character)
            if total_cell_offset > cell_width:
                return column_index
            column_index += 1

        # Account for the appearance of the tab character for this part
        total_cell_offset += expanded_tab_width
        # Check if the click falls within the boundary of the expanded tab.
        if total_cell_offset > cell_width:
            return column_index

        column_index += 1

    return len(line)



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import Static


class NestingDemo(App):
    """App with nested CSS."""

    CSS_PATH = "nesting02.tcss"

    def compose(self) -> ComposeResult:
        with Horizontal(id="questions"):
            yield Static("Yes", classes="button affirmative")
            yield Static("No", classes="button negative")


if __name__ == "__main__":
    app = NestingDemo()
    app.run()



from textual.app import App
from textual.containers import VerticalScroll
from textual.widgets import Label, Placeholder, Static


class Ruler(Static):
    def compose(self):
        ruler_text = "·\n·\n·\n·\n•\n" * 100
        yield Label(ruler_text)


class HeightComparisonApp(App):
    CSS_PATH = "height_comparison.tcss"

    def compose(self):
        yield VerticalScroll(
            Placeholder(id="cells"),  # (1)!
            Placeholder(id="percent"),
            Placeholder(id="w"),
            Placeholder(id="h"),
            Placeholder(id="vw"),
            Placeholder(id="vh"),
            Placeholder(id="auto"),
            Placeholder(id="fr1"),
            Placeholder(id="fr2"),
        )
        yield Ruler()


if __name__ == "__main__":
    app = HeightComparisonApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Label

COLORS = ("primary", "secondary", "accent", "warning", "error", "success")


class MutedBackgrounds(App[None]):
    CSS = "\n".join(
        f".text-{color} {{padding: 0 1; color: $text-{color}; background: ${color}-muted;}}"
        for color in COLORS
    )

    def compose(self) -> ComposeResult:
        for color in COLORS:
            yield Label(f"$text-{color} on ${color}-muted", classes=f"text-{color}")


app = MutedBackgrounds()
if __name__ == "__main__":
    app.run()



import pytest

from textual import on
from textual.app import App, ComposeResult
from textual.validation import Number, ValidationResult
from textual.widgets import Input


class InputApp(App):
    def __init__(self, validate_on=None):
        super().__init__()
        self.messages = []
        self.validator = Number(minimum=1, maximum=5)
        self.validate_on = validate_on

    def compose(self) -> ComposeResult:
        yield Input(
            validators=self.validator,
            validate_on=self.validate_on,
        )

    @on(Input.Changed)
    @on(Input.Submitted)
    def on_changed_or_submitted(self, event):
        self.messages.append(event)


async def test_input_changed_message_validation_failure():
    app = InputApp()
    async with app.run_test() as pilot:
        input = app.query_one(Input)
        input.value = "8"
        assert not input.is_valid
        await pilot.pause()
        assert len(app.messages) == 1
        assert app.messages[0].validation_result == ValidationResult.failure(
            failures=[
                Number.NotInRange(
                    value="8",
                    validator=app.validator,
                    description="Must be between 1 and 5.",
                )
            ],
        )


async def test_input_changed_message_validation_success():
    app = InputApp()
    async with app.run_test() as pilot:
        input = app.query_one(Input)
        input.value = "3"
        await pilot.pause()
        assert len(app.messages) == 1
        assert app.messages[0].validation_result == ValidationResult.success()


async def test_input_submitted_message_validation_failure():
    app = InputApp()
    async with app.run_test() as pilot:
        input = app.query_one(Input)
        input.value = "8"
        await input.action_submit()
        await pilot.pause()
        assert len(app.messages) == 2
        assert app.messages[1].validation_result == ValidationResult.failure(
            failures=[
                Number.NotInRange(
                    value="8",
                    validator=app.validator,
                    description="Must be between 1 and 5.",
                )
            ],
        )


async def test_input_submitted_message_validation_success():
    app = InputApp()
    async with app.run_test() as pilot:
        input = app.query_one(Input)
        input.value = "3"
        await input.action_submit()
        await pilot.pause()
        assert len(app.messages) == 2
        assert app.messages[1].validation_result == ValidationResult.success()


async def test_on_blur_triggers_validation():
    app = InputApp()
    async with app.run_test() as pilot:
        input = app.query_one(Input)
        input.focus()
        input.value = "3"
        input.remove_class("-valid")
        app.set_focus(None)
        await pilot.pause()
        assert input.has_class("-valid")


@pytest.mark.parametrize(
    "validate_on",
    [
        set(),
        [],
        {"blur"},
        ["submitted"],
        ["blur", "submitted"],
        ["fried", "garbage"],
    ],
)
async def test_validation_on_changed_should_not_happen(validate_on):
    app = InputApp(validate_on)
    async with app.run_test() as pilot:
        # sanity checks
        assert len(app.messages) == 0
        input = app.query_one(Input)
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")

        input.value = "3"
        await pilot.pause()
        assert len(app.messages) == 1
        assert app.messages[-1].validation_result is None
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")


@pytest.mark.parametrize(
    "validate_on",
    [
        set(),
        {"blur"},
        {"changed"},
        {"blur", "changed"},
        {"fried", "garbage"},
    ],
)
async def test_validation_on_submitted_should_not_happen(validate_on):
    app = InputApp(validate_on)
    async with app.run_test() as pilot:
        # sanity checks
        assert len(app.messages) == 0
        input = app.query_one(Input)
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")

        await input.action_submit()
        await pilot.pause()
        assert len(app.messages) == 1
        assert app.messages[-1].validation_result is None
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")


@pytest.mark.parametrize(
    "validate_on",
    [
        set(),
        {"submitted"},
        {"changed"},
        {"submitted", "changed"},
        {"fried", "garbage"},
    ],
)
async def test_validation_on_blur_should_not_happen_unless_specified(validate_on):
    app = InputApp(validate_on)
    async with app.run_test() as pilot:
        # sanity checks
        input = app.query_one(Input)
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")

        input.focus()
        await pilot.pause()
        app.set_focus(None)
        await pilot.pause()
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")


async def test_none_validate_on_means_all_validations_happen():
    app = InputApp(None)
    async with app.run_test() as pilot:
        assert len(app.messages) == 0  # sanity checks
        input = app.query_one(Input)
        assert not input.has_class("-valid")
        assert not input.has_class("-invalid")

        input.value = "3"
        await pilot.pause()
        assert len(app.messages) == 1
        assert app.messages[-1].validation_result is not None
        assert input.has_class("-valid")

        input.remove_class("-valid")

        await input.action_submit()
        await pilot.pause()
        assert len(app.messages) == 2
        assert app.messages[-1].validation_result is not None
        assert input.has_class("-valid")

        input.remove_class("-valid")

        input.focus()
        await pilot.pause()
        app.set_focus(None)
        await pilot.pause()
        assert input.has_class("-valid")


async def test_valid_empty():
    app = InputApp(None)
    async with app.run_test() as pilot:
        input = app.query_one(Input)

        await pilot.press("1", "backspace")

        assert not input.has_class("-valid")
        assert input.has_class("-invalid")

        input.valid_empty = True

        assert input.has_class("-valid")
        assert not input.has_class("-invalid")



from textual.app import App, ComposeResult
from textual.widgets import Footer, Label, Tabs

NAMES = [
    "Paul Atreidies",
    "Duke Leto Atreides",
    "Lady Jessica",
    "Gurney Halleck",
    "Baron Vladimir Harkonnen",
    "Glossu Rabban",
    "Chani",
    "Silgar",
]


class TabsApp(App):
    """Demonstrates the Tabs widget."""

    CSS = """
    Tabs {
        dock: top;
    }
    Screen {
        align: center middle;
    }
    Label {
        margin:1 1;
        width: 100%;
        height: 100%;
        background: $panel;
        border: tall $primary;
        content-align: center middle;
    }
    """

    BINDINGS = [
        ("a", "add", "Add tab"),
        ("r", "remove", "Remove active tab"),
        ("c", "clear", "Clear tabs"),
    ]

    def compose(self) -> ComposeResult:
        yield Tabs(NAMES[0])
        yield Label()
        yield Footer()

    def on_mount(self) -> None:
        """Focus the tabs when the app starts."""
        self.query_one(Tabs).focus()

    def on_tabs_tab_activated(self, event: Tabs.TabActivated) -> None:
        """Handle TabActivated message sent by Tabs."""
        label = self.query_one(Label)
        if event.tab is None:
            # When the tabs are cleared, event.tab will be None
            label.visible = False
        else:
            label.visible = True
            label.update(event.tab.label)

    def action_add(self) -> None:
        """Add a new tab."""
        tabs = self.query_one(Tabs)
        # Cycle the names
        NAMES[:] = [*NAMES[1:], NAMES[0]]
        tabs.add_tab(NAMES[0])

    def action_remove(self) -> None:
        """Remove active tab."""
        tabs = self.query_one(Tabs)
        active_tab = tabs.active_tab
        if active_tab is not None:
            tabs.remove_tab(active_tab.id)

    def action_clear(self) -> None:
        """Clear the tabs."""
        self.query_one(Tabs).clear()


if __name__ == "__main__":
    app = TabsApp()
    app.run()



from typing import List

from textual import on
from textual.app import App, ComposeResult
from textual.events import Event
from textual.message import Message
from textual.widgets import TextArea


class TextAreaApp(App):
    def __init__(self):
        super().__init__()
        self.messages = []

    @on(TextArea.Changed)
    @on(TextArea.SelectionChanged)
    def message_received(self, message: Message):
        self.messages.append(message)

    def compose(self) -> ComposeResult:
        yield TextArea("123")


def get_changed_messages(messages: List[Event]) -> List[TextArea.Changed]:
    return [message for message in messages if isinstance(message, TextArea.Changed)]


def get_selection_changed_messages(
    messages: List[Event],
) -> List[TextArea.SelectionChanged]:
    return [
        message
        for message in messages
        if isinstance(message, TextArea.SelectionChanged)
    ]


async def test_changed_message_edit_via_api():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        assert get_changed_messages(app.messages) == []

        text_area.insert("A")
        await pilot.pause()

        assert get_changed_messages(app.messages) == [TextArea.Changed(text_area)]
        assert get_selection_changed_messages(app.messages) == [
            TextArea.SelectionChanged(text_area.selection, text_area)
        ]


async def test_changed_message_via_typing():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        assert get_changed_messages(app.messages) == []

        await pilot.press("a")

        assert get_changed_messages(app.messages) == [TextArea.Changed(text_area)]
        assert get_selection_changed_messages(app.messages) == [
            TextArea.SelectionChanged(text_area.selection, text_area)
        ]


async def test_changed_message_edit_via_assignment():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        assert get_changed_messages(app.messages) == []

        text_area.text = ""
        await pilot.pause()

        assert get_changed_messages(app.messages) == [TextArea.Changed(text_area)]
        assert get_selection_changed_messages(app.messages) == []


async def test_selection_changed_via_api():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        assert get_selection_changed_messages(app.messages) == []

        text_area.cursor_location = (0, 1)
        await pilot.pause()

        assert get_selection_changed_messages(app.messages) == [
            TextArea.SelectionChanged(text_area.selection, text_area)
        ]


async def test_selection_changed_via_typing():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        assert get_selection_changed_messages(app.messages) == []

        await pilot.press("a")

        assert get_selection_changed_messages(app.messages) == [
            TextArea.SelectionChanged(text_area.selection, text_area)
        ]



from textual.app import App, ComposeResult
from textual.widgets import Button, TabbedContent


class FiddleWithTabsApp(App[None]):
    CSS = """
    TabPane:disabled {
        background: red;
    }
    """

    def compose(self) -> ComposeResult:
        with TabbedContent():
            yield Button()
            yield Button()
            yield Button()
            yield Button()
            yield Button()

    def on_mount(self) -> None:
        self.query_one(TabbedContent).disable_tab("tab-1")
        self.query_one(TabbedContent).disable_tab("tab-2")
        self.query_one(TabbedContent).hide_tab("tab-3")


if __name__ == "__main__":
    FiddleWithTabsApp().run()



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Header, Footer, Static, RadioSet
from textual.reactive import reactive


class Profile(Static):
    choices: reactive[list[str]] = reactive(list, recompose=True)

    def compose(self) -> ComposeResult:
        yield RadioSet(*self.choices)

    async def on_mount(self) -> None:
        self.choices.append("Foo")
        self.choices.append("Bar")
        self.mutate_reactive(Profile.choices)


class Landing(Screen):
    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Static(" Profile ", id="title")
        yield Profile()
        yield Footer()


class ForecastApp(App):
    """A Textual app to forecast Financials."""

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()

    def on_mount(self) -> None:
        self.install_screen(Landing(), name="landing")
        self.push_screen("landing")


if __name__ == "__main__":
    app = ForecastApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import DirectoryTree


class DirectoryTreeApp(App[None]):
    def compose(self) -> ComposeResult:
        tree = DirectoryTree(".")
        tree.show_root = True
        yield tree


async def test_managed_to_set_show_root_before_mounted() -> None:
    """https://github.com/Textualize/textual/issues/2363"""
    async with DirectoryTreeApp().run_test() as pilot:
        assert isinstance(pilot.app.query_one(DirectoryTree), DirectoryTree)
        assert pilot.app.query_one(DirectoryTree).show_root is True



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.reactive import reactive
from textual.widgets import Button, RichLog


class ValidateApp(App):
    CSS_PATH = "validate01.tcss"

    count = reactive(0)

    def validate_count(self, count: int) -> int:
        """Validate value."""
        if count < 0:
            count = 0
        elif count > 10:
            count = 10
        return count

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Button("+1", id="plus", variant="success"),
            Button("-1", id="minus", variant="error"),
            id="buttons",
        )
        yield RichLog(highlight=True)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "plus":
            self.count += 1
        else:
            self.count -= 1
        self.query_one(RichLog).write(f"count = {self.count}")


if __name__ == "__main__":
    app = ValidateApp()
    app.run()



import pytest
from pytest import approx
from rich.console import Console
from rich.text import Text

from textual.app import App
from textual.color import Gradient
from textual.css.query import NoMatches
from textual.renderables.bar import _apply_gradient
from textual.widget import Widget
from textual.widgets import ProgressBar


def test_initial_status():
    pb = ProgressBar()
    assert pb.total is None
    assert pb.progress == 0
    assert pb.percentage is None

    pb = ProgressBar(total=100)
    assert pb.total == 100
    assert pb.progress == 0
    assert pb.percentage == 0


def test_advance():
    pb = ProgressBar(total=100)

    pb.advance(10)
    assert pb.progress == 10
    assert pb.percentage == approx(0.1)

    pb.advance(42)
    assert pb.progress == 52
    assert pb.percentage == approx(0.52)

    pb.advance(0.0625)
    assert pb.progress == 52.0625
    assert pb.percentage == approx(0.520625)


def test_advance_backwards():
    pb = ProgressBar(total=100)

    pb.progress = 50

    pb.advance(-10)
    assert pb.progress == 40


def test_progress_overflow():
    pb = ProgressBar(total=100)

    pb.advance(999_999)
    assert pb.percentage == 1

    pb.update(total=50)
    assert pb.percentage == 1


def test_progress_underflow():
    pb = ProgressBar(total=100)

    pb.advance(-999_999)
    assert pb.percentage == 0


def test_non_negative_total():
    pb = ProgressBar(total=-100)
    assert pb.total == 0


def test_update_total():
    pb = ProgressBar()

    pb.update(total=100)
    assert pb.total == 100

    pb.update(total=1000)
    assert pb.total == 1000

    pb.update(total=None)
    assert pb.total is None

    pb.update(total=100)
    assert pb.total == 100


def test_update_progress():
    pb = ProgressBar(total=100)

    pb.update(progress=10)
    assert pb.progress == 10

    pb.update(progress=73)
    assert pb.progress == 73

    pb.update(progress=40)
    assert pb.progress == 40


def test_update_advance():
    pb = ProgressBar(total=100)

    pb.update(advance=10)
    assert pb.progress == 10

    pb.update(advance=10)
    assert pb.progress == 20

    pb.update(advance=10)
    assert pb.progress == 30


def test_update():
    pb = ProgressBar()

    pb.update(total=100, progress=30, advance=20)
    assert pb.total == 100
    assert pb.progress == 50


def test_go_back_to_indeterminate():
    pb = ProgressBar()

    pb.total = 100
    assert pb.percentage == 0
    pb.total = None
    assert pb.percentage is None


@pytest.mark.parametrize(
    ["show_bar", "show_percentage", "show_eta"],
    [
        (True, True, True),
        (True, True, False),
        (True, False, True),
        (True, False, False),
        (False, True, True),
        (False, True, False),
        (False, False, True),
        (False, False, False),
    ],
)
async def test_show_sub_widgets(show_bar: bool, show_percentage: bool, show_eta: bool):
    class PBApp(App[None]):
        def compose(self):
            self.pb = ProgressBar(
                show_bar=show_bar, show_percentage=show_percentage, show_eta=show_eta
            )
            yield self.pb

    app = PBApp()

    async with app.run_test():
        if show_bar:
            bar = app.pb.query_one("#bar")
            assert isinstance(bar, Widget)
        else:
            with pytest.raises(NoMatches):
                app.pb.query_one("#bar")

        if show_percentage:
            percentage = app.pb.query_one("#percentage")
            assert isinstance(percentage, Widget)
        else:
            with pytest.raises(NoMatches):
                app.pb.query_one("#percentage")

        if show_eta:
            eta = app.pb.query_one("#eta")
            assert isinstance(eta, Widget)
        else:
            with pytest.raises(NoMatches):
                app.pb.query_one("#eta")


def test_apply_gradient():
    text = Text("foo")
    gradient = Gradient.from_colors("red", "blue")
    _apply_gradient(text, gradient, 1)
    console = Console()
    assert text.get_style_at_offset(console, 0).color.triplet == (255, 0, 0)



from __future__ import annotations

from rich.table import Table

from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, OptionList

COLONIES: tuple[tuple[str, str, str, str], ...] = (
    ("Aerilon", "Demeter", "1.2 Billion", "Gaoth"),
    ("Aquaria", "Hermes", "75,000", "None"),
    ("Canceron", "Hephaestus", "6.7 Billion", "Hades"),
    ("Caprica", "Apollo", "4.9 Billion", "Caprica City"),
    ("Gemenon", "Hera", "2.8 Billion", "Oranu"),
    ("Leonis", "Artemis", "2.6 Billion", "Luminere"),
    ("Libran", "Athena", "2.1 Billion", "None"),
    ("Picon", "Poseidon", "1.4 Billion", "Queenstown"),
    ("Sagittaron", "Zeus", "1.7 Billion", "Tawa"),
    ("Scorpia", "Dionysus", "450 Million", "Celeste"),
    ("Tauron", "Ares", "2.5 Billion", "Hypatia"),
    ("Virgon", "Hestia", "4.3 Billion", "Boskirk"),
)


class OptionListApp(App[None]):
    CSS_PATH = "option_list.tcss"

    @staticmethod
    def colony(name: str, god: str, population: str, capital: str) -> Table:
        table = Table(title=f"Data for {name}", expand=True)
        table.add_column("Patron God")
        table.add_column("Population")
        table.add_column("Capital City")
        table.add_row(god, population, capital)
        return table

    def compose(self) -> ComposeResult:
        yield Header()
        yield OptionList(*[self.colony(*row) for row in COLONIES])
        yield Footer()


if __name__ == "__main__":
    OptionListApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, OptionList
from textual.widgets.option_list import Option


class OptionListApp(App[None]):
    CSS_PATH = "option_list.tcss"

    def compose(self) -> ComposeResult:
        yield Header()
        yield OptionList(
            Option("Aerilon", id="aer"),
            Option("Aquaria", id="aqu"),
            None,
            Option("Canceron", id="can"),
            Option("Caprica", id="cap", disabled=True),
            None,
            Option("Gemenon", id="gem"),
            None,
            Option("Leonis", id="leo"),
            Option("Libran", id="lib"),
            None,
            Option("Picon", id="pic"),
            None,
            Option("Sagittaron", id="sag"),
            Option("Scorpia", id="sco"),
            None,
            Option("Tauron", id="tau"),
            None,
            Option("Virgon", id="vir"),
        )
        yield Footer()


if __name__ == "__main__":
    OptionListApp().run()



from rich.text import Text

from textual.app import App, ComposeResult, RenderResult
from textual.containers import VerticalScroll
from textual.widgets import Header, Footer
from textual.widget import Widget


class Tester(Widget, can_focus=True):
    COMPONENT_CLASSES = {"tester--text"}

    DEFAULT_CSS = """
    Tester {
        height: auto;
    }
    
    Tester:focus > .tester--text {
        background: red;
    }
    """

    def __init__(self, n: int) -> None:
        self.n = n
        super().__init__()

    def render(self) -> RenderResult:
        return Text(
            f"test widget {self.n}", style=self.get_component_rich_style("tester--text")
        )


class StyleBugApp(App[None]):
    def compose(self) -> ComposeResult:
        yield Header()
        with VerticalScroll():
            for n in range(40):
                yield Tester(n)
        yield Footer()


if __name__ == "__main__":
    StyleBugApp().run()



from textual import on, work
from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Button, Label


class QuestionScreen(Screen[bool]):
    """Screen with a parameter."""

    def __init__(self, question: str) -> None:
        self.question = question
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label(self.question)
        yield Button("Yes", id="yes", variant="success")
        yield Button("No", id="no")

    @on(Button.Pressed, "#yes")
    def handle_yes(self) -> None:
        self.dismiss(True)  # (1)!

    @on(Button.Pressed, "#no")
    def handle_no(self) -> None:
        self.dismiss(False)  # (2)!


class QuestionsApp(App):
    """Demonstrates wait_for_dismiss"""

    CSS_PATH = "questions01.tcss"

    @work  # (3)!
    async def on_mount(self) -> None:
        if await self.push_screen_wait(  # (4)!
            QuestionScreen("Do you like Textual?"),
        ):
            self.notify("Good answer!")
        else:
            self.notify(":-(", severity="error")


if __name__ == "__main__":
    app = QuestionsApp()
    app.run()



"""
Fuzzy matcher.

This class is used by the [command palette](/guide/command_palette) to match search terms.

"""

from __future__ import annotations

from operator import itemgetter
from re import IGNORECASE, escape, finditer, search
from typing import Iterable, NamedTuple

import rich.repr

from textual.content import Content
from textual.visual import Style


class _Search(NamedTuple):
    """Internal structure to keep track of a recursive search."""

    candidate_offset: int = 0
    query_offset: int = 0
    offsets: tuple[int, ...] = ()

    def branch(self, offset: int) -> tuple[_Search, _Search]:
        """Branch this search when an offset is found.

        Args:
            offset: Offset of a matching letter in the query.

        Returns:
            A pair of search objects.
        """
        _, query_offset, offsets = self
        return (
            _Search(offset + 1, query_offset + 1, offsets + (offset,)),
            _Search(offset + 1, query_offset, offsets),
        )

    @property
    def groups(self) -> int:
        """Number of groups in offsets."""
        groups = 1
        last_offset = self.offsets[0]
        for offset in self.offsets[1:]:
            if offset != last_offset + 1:
                groups += 1
            last_offset = offset
        return groups


class FuzzySearch:
    """Performs a fuzzy search.

    Unlike a regex solution, this will finds all possible matches.
    """

    def __init__(self, case_sensitive: bool = False) -> None:
        """Initialize fuzzy search.

        Args:
            case_sensitive: Is the match case sensitive?
        """
        self.cache: dict[tuple[str, str, bool], tuple[float, tuple[int, ...]]] = {}
        self.case_sensitive = case_sensitive

    def match(self, query: str, candidate: str) -> tuple[float, tuple[int, ...]]:
        """Match against a query.

        Args:
            query: The fuzzy query.
            candidate: A candidate to check,.

        Returns:
            A pair of (score, tuple of offsets). `(0, ())` for no result.
        """

        query_regex = ".*?".join(f"({escape(character)})" for character in query)
        if not search(
            query_regex, candidate, flags=0 if self.case_sensitive else IGNORECASE
        ):
            # Bail out early if there is no possibility of a match
            return (0.0, ())

        cache_key = (query, candidate, self.case_sensitive)
        if cache_key in self.cache:
            return self.cache[cache_key]
        result = max(
            self._match(query, candidate), key=itemgetter(0), default=(0.0, ())
        )
        self.cache[cache_key] = result
        return result

    def _match(
        self, query: str, candidate: str
    ) -> Iterable[tuple[float, tuple[int, ...]]]:
        """Generator to do the matching.

        Args:
            query: Query to match.
            candidate: Candidate to check against.

        Yields:
            Pairs of score and tuple of offsets.
        """
        if not self.case_sensitive:
            query = query.lower()
            candidate = candidate.lower()

        # We need this to give a bonus to first letters.
        first_letters = {match.start() for match in finditer(r"\w+", candidate)}

        def score(search: _Search) -> float:
            """Sore a search.

            Args:
                search: Search object.

            Returns:
                Score.

            """
            # This is a heuristic, and can be tweaked for better results
            # Boost first letter matches
            score: float = sum(
                (2.0 if offset in first_letters else 1.0) for offset in search.offsets
            )
            # Boost to favor less groups
            offset_count = len(search.offsets)
            normalized_groups = (offset_count - (search.groups - 1)) / offset_count
            score *= 1 + (normalized_groups**2)
            return score

        stack: list[_Search] = [_Search()]
        push = stack.append
        pop = stack.pop
        query_size = len(query)
        find = candidate.find
        # Limit the number of loops out of an abundance of caution.
        # This would be hard to reach without contrived data.
        remaining_loops = 200

        while stack and (remaining_loops := remaining_loops - 1):
            search = pop()
            offset = find(query[search.query_offset], search.candidate_offset)
            if offset != -1:
                advance_branch, branch = search.branch(offset)
                if advance_branch.query_offset == query_size:
                    yield score(advance_branch), advance_branch.offsets
                    push(branch)
                else:
                    push(advance_branch)
                    push(branch)


@rich.repr.auto
class Matcher:
    """A fuzzy matcher."""

    def __init__(
        self,
        query: str,
        *,
        match_style: Style | None = None,
        case_sensitive: bool = False,
    ) -> None:
        """Initialise the fuzzy matching object.

        Args:
            query: A query as typed in by the user.
            match_style: The style to use to highlight matched portions of a string.
            case_sensitive: Should matching be case sensitive?
        """
        self._query = query
        self._match_style = Style(reverse=True) if match_style is None else match_style
        self._case_sensitive = case_sensitive
        self.fuzzy_search = FuzzySearch()

    @property
    def query(self) -> str:
        """The query string to look for."""
        return self._query

    @property
    def match_style(self) -> Style:
        """The style that will be used to highlight hits in the matched text."""
        return self._match_style

    @property
    def case_sensitive(self) -> bool:
        """Is this matcher case sensitive?"""
        return self._case_sensitive

    def match(self, candidate: str) -> float:
        """Match the candidate against the query.

        Args:
            candidate: Candidate string to match against the query.

        Returns:
            Strength of the match from 0 to 1.
        """
        return self.fuzzy_search.match(self.query, candidate)[0]

    def highlight(self, candidate: str) -> Content:
        """Highlight the candidate with the fuzzy match.

        Args:
            candidate: The candidate string to match against the query.

        Returns:
            A [rich.text.Text][`Text`] object with highlighted matches.
        """
        content = Content.from_markup(candidate)
        score, offsets = self.fuzzy_search.match(self.query, candidate)
        if not score:
            return content
        for offset in offsets:
            if not candidate[offset].isspace():
                content = content.stylize(self._match_style, offset, offset + 1)
        return content



"""The standard path of selecting text on focus is well covered by snapshot tests."""

from textual import events
from textual.app import App, ComposeResult
from textual.widgets import Input
from textual.widgets.input import Selection


class InputApp(App[None]):
    """An app with an input widget."""

    def compose(self) -> ComposeResult:
        yield Input("Hello, world!")


async def test_focus_from_app_focus_does_not_select():
    """When an Input has focused and the *app* is blurred and then focused (e.g. by pressing
    alt+tab or focusing another terminal pane), then the content of the Input should not be
    fully selected when `Input.select_on_focus=True`.
    """
    async with InputApp().run_test() as pilot:
        input_widget = pilot.app.query_one(Input)
        input_widget.focus()
        input_widget.selection = Selection.cursor(0)
        assert input_widget.selection == Selection.cursor(0)
        pilot.app.post_message(events.AppBlur())
        await pilot.pause()
        pilot.app.post_message(events.AppFocus())
        await pilot.pause()
        assert input_widget.selection == Selection.cursor(0)



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Placeholder


class VerticalApp(App):
    CSS = """
    #top {
        min-height: 10;        
        height: 1fr;
        border: white;
    }
    #bottom {
        height:3fr;
        border: white;
    }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Placeholder(id="top")
            yield Placeholder(id="bottom")


if __name__ == "__main__":
    VerticalApp().run()



from textual.app import App
from textual.containers import Grid
from textual.widgets import Label


class MyApp(App):
    CSS_PATH = "grid_size_columns.tcss"

    def compose(self):
        yield Grid(
            Label("1"),
            Label("2"),
            Label("3"),
            Label("4"),
            Label("5"),
        )


if __name__ == "__main__":
    app = MyApp()
    app.run()



from __future__ import annotations

import re
from typing import TYPE_CHECKING, NamedTuple

import rich.repr
from rich.console import Group, RenderableType
from rich.highlighter import ReprHighlighter
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from textual.css._error_tools import friendly_list
from textual.css.constants import VALID_PSEUDO_CLASSES
from textual.suggestions import get_suggestion

if TYPE_CHECKING:
    from textual.css.types import CSSLocation


class TokenError(Exception):
    """Error raised when the CSS cannot be tokenized (syntax error)."""

    def __init__(
        self,
        read_from: CSSLocation,
        code: str,
        start: tuple[int, int],
        message: str,
        end: tuple[int, int] | None = None,
    ) -> None:
        """
        Args:
            read_from: The location where the CSS was read from.
            code: The code being parsed.
            start: Line and column number of the error (1-indexed).
            message: A message associated with the error.
            end: End location of token (1-indexed), or None if not known.
        """

        self.read_from = read_from
        self.code = code
        self.start = start
        self.end = end or start
        super().__init__(message)

    def _get_snippet(self) -> Panel:
        """Get a short snippet of code around a given line number.

        Returns:
            A renderable.
        """
        from rich.syntax import Syntax

        line_no = self.start[0]
        # TODO: Highlight column number
        syntax = Syntax(
            self.code,
            lexer="scss",
            theme="ansi_light",
            line_numbers=True,
            indent_guides=True,
            line_range=(max(0, line_no - 2), line_no + 2),
            highlight_lines={line_no},
        )
        syntax.stylize_range(
            "reverse bold",
            (self.start[0], self.start[1] - 1),
            (self.end[0], self.end[1] - 1),
        )
        return Panel(syntax, border_style="red")

    def __rich__(self) -> RenderableType:
        highlighter = ReprHighlighter()
        errors: list[RenderableType] = []

        message = str(self)
        errors.append(Text(" Error in stylesheet:", style="bold red"))

        line_no, col_no = self.start

        path, widget_variable = self.read_from
        if widget_variable:
            css_location = f" {path}, {widget_variable}:{line_no}:{col_no}"
        else:
            css_location = f" {path}:{line_no}:{col_no}"
        errors.append(highlighter(css_location))
        errors.append(self._get_snippet())

        final_message = "\n".join(
            f"• {message_part.strip()}" for message_part in message.split(";")
        )
        errors.append(
            Padding(
                highlighter(
                    Text(final_message, "red"),
                ),
                pad=(0, 1),
            )
        )

        return Group(*errors)


class EOFError(TokenError):
    """Indicates that the CSS ended prematurely."""


@rich.repr.auto
class Expect:
    """Object that describes the format of tokens."""

    def __init__(self, description: str, **tokens: str) -> None:
        """Create Expect object.

        Args:
            description: Description of this class of tokens, used in errors.
        """
        self.description = f"Expected {description}"
        self.names = list(tokens.keys())
        self.regexes = list(tokens.values())
        self._regex = re.compile(
            "("
            + "|".join(f"(?P<{name}>{regex})" for name, regex in tokens.items())
            + ")"
        )
        self.match = self._regex.match
        self.search = self._regex.search
        self._expect_eof = False
        self._expect_semicolon = True
        self._extract_text = False

    def expect_eof(self, eof: bool = True) -> Expect:
        """Expect an end of file."""
        self._expect_eof = eof
        return self

    def expect_semicolon(self, semicolon: bool = True) -> Expect:
        """Tokenizer expects text to be terminated with a semi-colon."""
        self._expect_semicolon = semicolon
        return self

    def extract_text(self, extract: bool = True) -> Expect:
        self._extract_text = extract
        return self

    def __rich_repr__(self) -> rich.repr.Result:
        yield from zip(self.names, self.regexes)


class ReferencedBy(NamedTuple):
    name: str
    location: tuple[int, int]
    length: int
    code: str


@rich.repr.auto(angular=True)
class Token(NamedTuple):
    name: str
    value: str
    read_from: CSSLocation
    code: str
    location: tuple[int, int]
    """Token starting location, 0-indexed."""
    referenced_by: ReferencedBy | None = None

    @property
    def start(self) -> tuple[int, int]:
        """Start line and column (1-indexed)."""
        line, offset = self.location
        return (line + 1, offset + 1)

    @property
    def end(self) -> tuple[int, int]:
        """End line and column (1-indexed)."""
        line, offset = self.location
        return (line + 1, offset + len(self.value) + 1)

    def with_reference(self, by: ReferencedBy | None) -> "Token":
        """Return a copy of the Token, with reference information attached.
        This is used for variable substitution, where a variable reference
        can refer to tokens which were defined elsewhere. With the additional
        ReferencedBy data attached, we can track where the token we are referring
        to is used.
        """
        return Token(
            name=self.name,
            value=self.value,
            read_from=self.read_from,
            code=self.code,
            location=self.location,
            referenced_by=by,
        )

    def __str__(self) -> str:
        return self.value

    def __rich_repr__(self) -> rich.repr.Result:
        yield "name", self.name
        yield "value", self.value
        yield (
            "read_from",
            self.read_from[0] if not self.read_from[1] else self.read_from,
        )
        yield "code", self.code if len(self.code) < 40 else self.code[:40] + "..."
        yield "location", self.location
        yield "referenced_by", self.referenced_by, None


class Tokenizer:
    """Tokenizes Textual CSS."""

    def __init__(self, text: str, read_from: CSSLocation = ("", "")) -> None:
        """Initialize the tokenizer.

        Args:
            text: String containing CSS.
            read_from: Information regarding where the CSS was read from.
        """
        self.read_from = read_from
        self.code = text
        self.lines = text.splitlines(keepends=True)
        self.line_no = 0
        self.col_no = 0

    def get_token(self, expect: Expect) -> Token:
        """Get the next token.

        Args:
            expect: Expect object which describes which tokens may be read.

        Raises:
            EOFError: If there is an unexpected end of file.
            TokenError: If there is an error with the token.

        Returns:
            A new Token.
        """

        line_no = self.line_no
        col_no = self.col_no
        if line_no >= len(self.lines):
            if expect._expect_eof:
                return Token(
                    "eof",
                    "",
                    self.read_from,
                    self.code,
                    (line_no, col_no),
                    None,
                )
            else:
                raise EOFError(
                    self.read_from,
                    self.code,
                    (line_no + 1, col_no + 1),
                    "Unexpected end of file; did you forget a '}' ?",
                )
        line = self.lines[line_no]
        preceding_text: str = ""
        if expect._extract_text:
            match = expect.search(line, col_no)
            if match is None:
                preceding_text = line[self.col_no :]
                self.line_no += 1
                self.col_no = 0
            else:
                col_no = match.start()
                preceding_text = line[self.col_no : col_no]
                self.col_no = col_no
            if preceding_text:
                token = Token(
                    "text",
                    preceding_text,
                    self.read_from,
                    self.code,
                    (line_no, col_no),
                    referenced_by=None,
                )

                return token

        else:
            match = expect.match(line, col_no)

        if match is None:
            error_line = line[col_no:]
            error_message = (
                f"{expect.description} (found {error_line.split(';')[0]!r})."
            )
            if expect._expect_semicolon and not error_line.endswith(";"):
                error_message += "; Did you forget a semicolon at the end of a line?"
            raise TokenError(
                self.read_from, self.code, (line_no + 1, col_no + 1), error_message
            )

        for name, value in zip(expect.names, match.groups()[1:]):
            if value is not None:
                break
        else:
            # For MyPy's benefit
            raise AssertionError("can't reach here")

        token = Token(
            name,
            value,
            self.read_from,
            self.code,
            (line_no, col_no),
            referenced_by=None,
        )

        if (
            token.name == "pseudo_class"
            and token.value.strip(":") not in VALID_PSEUDO_CLASSES
        ):
            pseudo_class = token.value.strip(":")
            suggestion = get_suggestion(pseudo_class, list(VALID_PSEUDO_CLASSES))
            all_valid = f"must be one of {friendly_list(VALID_PSEUDO_CLASSES)}"
            if suggestion:
                raise TokenError(
                    self.read_from,
                    self.code,
                    (line_no + 1, col_no + 1),
                    f"unknown pseudo-class {pseudo_class!r}; did you mean {suggestion!r}?; {all_valid}",
                )
            else:
                raise TokenError(
                    self.read_from,
                    self.code,
                    (line_no + 1, col_no + 1),
                    f"unknown pseudo-class {pseudo_class!r}; {all_valid}",
                )

        col_no += len(value)
        if col_no >= len(line):
            line_no += 1
            col_no = 0
        self.line_no = line_no
        self.col_no = col_no
        return token

    def skip_to(self, expect: Expect) -> Token:
        """Skip tokens.

        Args:
            expect: Expect object describing the expected token.

        Raises:
            EOFError: If end of file is reached.

        Returns:
            A new token.
        """
        line_no = self.line_no
        col_no = self.col_no

        while True:
            if line_no >= len(self.lines):
                raise EOFError(
                    self.read_from,
                    self.code,
                    (line_no, col_no),
                    "Unexpected end of file; did you forget a '}' ?",
                )
            line = self.lines[line_no]
            match = expect.search(line, col_no)

            if match is None:
                line_no += 1
                col_no = 0
            else:
                self.line_no = line_no
                self.col_no = match.span(0)[0]
                return self.get_token(expect)



from textual.app import App, ComposeResult
from textual.widgets import Static
from textual.containers import Horizontal, Vertical, Grid


class Box(Static):
    pass


class KeylineApp(App):
    CSS = """
    Vertical {
        keyline: thin red;
    }
    Horizontal {
        keyline: heavy green;
    }
    Grid {
        keyline: double magenta;
    }
    Box {
        width: 1fr;
        height: 1fr;        
    }
    Horizontal > Box, Vertical > Box {
        margin: 1;
    }
    Grid {
        grid-size: 2;
        grid-gutter: 1;
    }

    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Box("1")
            yield Box("2")
            yield Box("3")
        with Horizontal():
            yield Box("4")
            yield Box("5")
            yield Box("6")
        with Grid():
            yield Box("7")
            yield Box("8")
            yield Box("9")


if __name__ == "__main__":
    app = KeylineApp()
    app.run()



from __future__ import annotations

import pytest

from textual._widget_navigation import (
    find_first_enabled,
    find_last_enabled,
    find_next_enabled,
    find_next_enabled_no_wrap,
    get_directed_distance,
)


class _D:
    def __init__(self, disabled):
        self.disabled = disabled


# Represent disabled/enabled objects that are compact to write in tests.
D = _D(True)
E = _D(False)


@pytest.mark.parametrize(
    ["index", "start", "direction", "wrap_at", "dist"],
    [
        (2, 8, 1, 10, 4),
        (2, 8, -1, 10, 6),
        (8, 2, -1, 10, 4),
        (8, 2, 1, 10, 6),
        (8, 2, 1, 1234123512, 6),
        (2, 8, 1, 11, 5),
        (2, 8, 1, 12, 6),
        (5, 5, 1, 10, 0),
    ],
)
def test_distance(index, start, direction, wrap_at, dist):
    assert (
        get_directed_distance(
            index=index,
            start=start,
            direction=direction,
            wrap_at=wrap_at,
        )
        == dist
    )


@pytest.mark.parametrize(
    "function",
    [
        find_first_enabled,
        find_last_enabled,
    ],
)
def test_find_enabled_returns_none_on_empty(function):
    assert function([]) is None


@pytest.mark.parametrize(
    ["candidates", "anchor", "direction", "result"],
    [
        # No anchor & no candidates -> no next
        ([], None, 1, None),
        ([], None, -1, None),
        # No anchor but candidates -> get first/last one
        ([E], None, 1, 0),
        ([E, D], None, 1, 0),
        ([E, E], None, 1, 0),
        ([D, E], None, 1, 1),
        ([D, D, E], None, 1, 2),
        ([E], None, -1, 0),
        ([E, E], None, -1, 1),
        ([E, D], None, -1, 0),
        ([E, D, D], None, -1, 0),
        # No enabled candidates -> return the anchor
        ([D, D, D], 0, 1, 0),
        ([D, D, D], 1, 1, 1),
        ([D, D, D], 1, -1, 1),
        ([D, D, D], None, -1, None),
        # General case
        # 0  1  2  3  4  5
        ([E, D, D, E, E, D], 0, 1, 3),
        ([E, D, D, E, E, D], 0, -1, 4),
        ([E, D, D, E, E, D], 1, 1, 3),
        ([E, D, D, E, E, D], 1, -1, 0),
        ([E, D, D, E, E, D], 2, 1, 3),
        ([E, D, D, E, E, D], 2, -1, 0),
        ([E, D, D, E, E, D], 3, 1, 4),
        ([E, D, D, E, E, D], 3, -1, 0),
        ([E, D, D, E, E, D], 4, 1, 0),
        ([E, D, D, E, E, D], 4, -1, 3),
        ([E, D, D, E, E, D], 5, 1, 0),
        ([E, D, D, E, E, D], 5, -1, 4),
    ],
)
def test_find_next_enabled(candidates, anchor, direction, result):
    assert find_next_enabled(candidates, anchor, direction) == result


@pytest.mark.parametrize(
    ["candidates", "anchor", "direction", "result"],
    [
        # No anchor & no candidates -> no next
        ([], None, 1, None),
        ([], None, -1, None),
        # No anchor but candidates -> get first/last one
        ([E], None, 1, 0),
        ([E, D], None, 1, 0),
        ([E, E], None, 1, 0),
        ([D, E], None, 1, 1),
        ([D, D, E], None, 1, 2),
        ([E], None, -1, 0),
        ([E, E], None, -1, 1),
        ([E, D], None, -1, 0),
        ([E, D, D], None, -1, 0),
        # No enabled candidates -> return None
        ([D, D, D], 0, 1, None),
        ([D, D, D], 1, 1, None),
        ([D, D, D], 1, -1, None),
        ([D, D, D], None, -1, None),
        # General case
        # 0  1  2  3  4  5
        ([E, D, D, E, E, D], 0, 1, 3),
        ([E, D, D, E, E, D], 0, -1, None),
        ([E, D, D, E, E, D], 1, 1, 3),
        ([E, D, D, E, E, D], 1, -1, 0),
        ([E, D, D, E, E, D], 2, 1, 3),
        ([E, D, D, E, E, D], 2, -1, 0),
        ([E, D, D, E, E, D], 3, 1, 4),
        ([E, D, D, E, E, D], 3, -1, 0),
        ([E, D, D, E, E, D], 4, 1, None),
        ([E, D, D, E, E, D], 4, -1, 3),
        ([E, D, D, E, E, D], 5, 1, None),
        ([E, D, D, E, E, D], 5, -1, 4),
    ],
)
def test_find_next_enabled_no_wrap(candidates, anchor, direction, result):
    assert find_next_enabled_no_wrap(candidates, anchor, direction) == result


@pytest.mark.parametrize(
    ["function", "start", "direction"],
    [
        (find_next_enabled_no_wrap, 0, 1),
        (find_next_enabled_no_wrap, 0, -1),
        (find_next_enabled_no_wrap, 1, 1),
        (find_next_enabled_no_wrap, 1, -1),
        (find_next_enabled_no_wrap, 2, 1),
        (find_next_enabled_no_wrap, 2, -1),
    ],
)
def test_find_next_with_anchor(function, start, direction):
    assert function([E, E, E], start, direction, True) == start



from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, SelectionList


class SelectionListApp(App[None]):
    CSS_PATH = "selection_list.tcss"

    def compose(self) -> ComposeResult:
        yield Header()
        yield SelectionList[int](  # (1)!
            ("Falken's Maze", 0, True),
            ("Black Jack", 1),
            ("Gin Rummy", 2),
            ("Hearts", 3),
            ("Bridge", 4),
            ("Checkers", 5),
            ("Chess", 6, True),
            ("Poker", 7),
            ("Fighter Combat", 8, True),
        )
        yield Footer()

    def on_mount(self) -> None:
        self.query_one(SelectionList).border_title = "Shall we play some games?"


if __name__ == "__main__":
    SelectionListApp().run()



from __future__ import annotations

from typing import Any

from textual.app import App, ComposeResult
from textual.binding import Binding, Keymap
from textual.dom import DOMNode
from textual.widget import Widget
from textual.widgets import Label


class Counter(App[None]):
    BINDINGS = [
        Binding(key="i,up", action="increment", id="app.increment"),
        Binding(key="d,down", action="decrement", id="app.decrement"),
    ]

    def __init__(self, keymap: Keymap, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.count = 0
        self.clashed_bindings: set[Binding] | None = None
        self.clashed_node: DOMNode | None = None
        self.keymap = keymap

    def compose(self) -> ComposeResult:
        yield Label("foo")

    def on_mount(self) -> None:
        self.set_keymap(self.keymap)

    def action_increment(self) -> None:
        self.count += 1

    def action_decrement(self) -> None:
        self.count -= 1

    def handle_bindings_clash(
        self, clashed_bindings: set[Binding], node: DOMNode
    ) -> None:
        self.clashed_bindings = clashed_bindings
        self.clashed_node = node


async def test_keymap_default_binding_replaces_old_binding():
    app = Counter({"app.increment": "right,k"})
    async with app.run_test() as pilot:
        # The original bindings are removed - action not called.
        await pilot.press("i", "up")
        assert app.count == 0

        # The new bindings are active and call the action.
        await pilot.press("right", "k")
        assert app.count == 2


async def test_keymap_sends_message_when_clash():
    app = Counter({"app.increment": "d"})
    async with app.run_test() as pilot:
        await pilot.press("d")
        assert app.clashed_bindings is not None
        assert len(app.clashed_bindings) == 1
        clash = app.clashed_bindings.pop()
        assert app.clashed_node is app
        assert clash.key == "d"
        assert clash.action == "increment"
        assert clash.id == "app.increment"


async def test_keymap_with_unknown_id_is_noop():
    app = Counter({"this.is.an.unknown.id": "d"})
    async with app.run_test() as pilot:
        await pilot.press("d")
        assert app.count == -1


async def test_keymap_inherited_bindings_same_id():
    """When a child widget inherits from a parent widget, if they have
    a binding with the same ID, then both parent and child bindings will
    be overridden by the keymap (assuming the keymap has a mapping with the
    same ID)."""

    parent_counter = 0
    child_counter = 0

    class Parent(Widget, can_focus=True):
        BINDINGS = [
            Binding(key="x", action="increment", id="increment"),
        ]

        def action_increment(self) -> None:
            nonlocal parent_counter
            parent_counter += 1

    class Child(Parent):
        BINDINGS = [
            Binding(key="x", action="increment", id="increment"),
        ]

        def action_increment(self) -> None:
            nonlocal child_counter
            child_counter += 1

    class MyApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Parent()
            yield Child()

        def on_mount(self) -> None:
            self.set_keymap({"increment": "i"})

    app = MyApp()
    async with app.run_test() as pilot:
        # Default binding is unbound due to keymap.
        await pilot.press("x")
        assert parent_counter == 0
        assert child_counter == 0

        # New binding is active, parent is focused - action called.
        await pilot.press("i")
        assert parent_counter == 1
        assert child_counter == 0

        # Tab to focus the child.
        await pilot.press("tab")

        # Default binding results in no change.
        await pilot.press("x")
        assert parent_counter == 1
        assert child_counter == 0

        # New binding is active, child is focused - action called.
        await pilot.press("i")
        assert parent_counter == 1
        assert child_counter == 1


async def test_keymap_child_with_different_id_overridden():
    """Ensures that overriding a parent binding doesn't influence a child
    binding with a different ID."""

    parent_counter = 0
    child_counter = 0

    class Parent(Widget, can_focus=True):
        BINDINGS = [
            Binding(key="x", action="increment", id="parent.increment"),
        ]

        def action_increment(self) -> None:
            nonlocal parent_counter
            parent_counter += 1

    class Child(Parent):
        BINDINGS = [
            Binding(key="x", action="increment", id="child.increment"),
        ]

        def action_increment(self) -> None:
            nonlocal child_counter
            child_counter += 1

    class MyApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Parent()
            yield Child()

        def on_mount(self) -> None:
            self.set_keymap({"parent.increment": "i"})

    app = MyApp()
    async with app.run_test() as pilot:
        # Default binding is unbound due to keymap.
        await pilot.press("x")
        assert parent_counter == 0
        assert child_counter == 0

        # New binding is active, parent is focused - action called.
        await pilot.press("i")
        assert parent_counter == 1
        assert child_counter == 0

        # Tab to focus the child.
        await pilot.press("tab")

        # Default binding is still active on the child.
        await pilot.press("x")
        assert parent_counter == 1
        assert child_counter == 1

        # The binding from the keymap only affects the parent, so
        # pressing it with the child focused does nothing.
        await pilot.press("i")
        assert parent_counter == 1
        assert child_counter == 1



"""
A simple animated sidebar.

See comments for details.

"""

from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Footer, Label

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain."""


class Sidebar(Widget):
    """
    Our sidebar widget.

    Add desired content to compose()

    """

    DEFAULT_CSS = """
    Sidebar {
        width: 30;
        /* Needs to go in its own layer to sit above content */
        layer: sidebar; 
        /* Dock the sidebar to the appropriate side */
        dock: left;
        /* Offset x to be -100% to move it out of view by default */
        offset-x: -100%;

        background: $primary;
        border-right: vkey $background;    

        /* Enable animation */
        transition: offset 200ms;
        
        &.-visible {
            /* Set offset.x to 0 to make it visible when class is applied */
            offset-x: 0;
        }

        & > Vertical {
            margin: 1 2;
        }
    }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("Your sidebar here!")


class SidebarApp(App):
    """
    Test app to show our sidebar.
    """

    DEFAULT_CSS = """
    Screen {    
        layers: sidebar;
    }

    """

    BINDINGS = [("s", "toggle_sidebar", "Toggle Sidebar")]

    show_sidebar = reactive(False)

    def compose(self) -> ComposeResult:
        yield Sidebar()
        yield Label(TEXT)
        yield Footer()

    def action_toggle_sidebar(self) -> None:
        """Toggle the sidebar visibility."""
        self.show_sidebar = not self.show_sidebar

    def watch_show_sidebar(self, show_sidebar: bool) -> None:
        """Set or unset visible class when reactive changes."""
        self.query_one(Sidebar).set_class(show_sidebar, "-visible")


if __name__ == "__main__":
    app = SidebarApp()
    app.run()



"""
General exception classes.

"""

from __future__ import annotations


class TextualError(Exception):
    """Base class for Textual errors."""


class NoWidget(TextualError):
    """Specified widget was not found."""


class RenderError(TextualError):
    """An object could not be rendered."""


class DuplicateKeyHandlers(TextualError):
    """More than one handler for a single key press.

    For example, if the handlers `key_ctrl_i` and `key_tab` were defined on the same
    widget, then this error would be raised.
    """



from pathlib import Path

from textual.app import App, ComposeResult
from textual.widgets import DirectoryTree


class DirectoryTreeReloadApp(App[None]):
    BINDINGS = [
        ("r", "reload"),
        ("e", "expand"),
        ("d", "delete"),
    ]

    async def setup(self, path_root: Path) -> None:
        self.path_root = path_root

        structure = [
            "f1.txt",
            "f2.txt",
            "b1/f1.txt",
            "b1/f2.txt",
            "b2/f1.txt",
            "b2/f2.txt",
            "b1/c1/f1.txt",
            "b1/c1/f2.txt",
            "b1/c2/f1.txt",
            "b1/c2/f2.txt",
            "b1/c1/d1/f1.txt",
            "b1/c1/d1/f2.txt",
            "b1/c1/d2/f1.txt",
            "b1/c1/d2/f2.txt",
        ]
        for file in structure:
            path = path_root / Path(file)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch(exist_ok=True)

        await self.mount(DirectoryTree(self.path_root))

    async def action_reload(self) -> None:
        dt = self.query_one(DirectoryTree)
        await dt.reload()

    def action_expand(self) -> None:
        self.query_one(DirectoryTree).root.expand_all()

    def action_delete(self) -> None:
        self.rmdir(self.path_root / Path("b1/c1/d2"))
        self.rmdir(self.path_root / Path("b1/c2"))
        self.rmdir(self.path_root / Path("b2"))

    def rmdir(self, path: Path) -> None:
        for file in path.iterdir():
            if file.is_file():
                file.unlink()
            elif file.is_dir():
                self.rmdir(file)
        path.rmdir()



from textual.content import Span
from textual.fuzzy import Matcher
from textual.style import Style


def test_no_match():
    """Check non matching score of zero."""
    matcher = Matcher("x")
    assert matcher.match("foo") == 0


def test_match_single_group():
    """Check that single groups rang higher."""
    matcher = Matcher("abc")
    assert matcher.match("foo abc bar") > matcher.match("fooa barc")


def test_boosted_matches():
    """Check first word matchers rank higher."""
    matcher = Matcher("ss")

    # First word matchers should score higher
    assert matcher.match("Save Screenshot") > matcher.match("Show Keys abcde")


def test_highlight():
    matcher = Matcher("foo.bar")

    spans = matcher.highlight("foo/egg.bar").spans
    print(repr(spans))
    assert spans == [
        Span(0, 1, Style(reverse=True)),
        Span(1, 2, Style(reverse=True)),
        Span(2, 3, Style(reverse=True)),
        Span(7, 8, Style(reverse=True)),
        Span(8, 9, Style(reverse=True)),
        Span(9, 10, Style(reverse=True)),
        Span(10, 11, Style(reverse=True)),
    ]



"""
Tests for the switch toggle animation, which is considered a basic animation.
(An animation that also plays on the level BASIC.)
"""

from textual.app import App, ComposeResult
from textual.widgets import Switch


class SwitchApp(App[None]):
    def compose(self) -> ComposeResult:
        yield Switch()


async def test_switch_animates_on_full() -> None:
    app = SwitchApp()
    app.animation_level = "full"

    async with app.run_test() as pilot:
        switch = app.query_one(Switch)
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        switch.action_toggle_switch()
        await pilot.pause()
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        # Move to the next frame.
        animator()
        # The animation should still be running.
        assert app.animator.is_being_animated(switch, "_slider_position")


async def test_switch_animates_on_basic() -> None:
    app = SwitchApp()
    app.animation_level = "basic"

    async with app.run_test() as pilot:
        switch = app.query_one(Switch)
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        switch.action_toggle_switch()
        await pilot.pause()
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        # Move to the next frame.
        animator()
        # The animation should still be running.
        assert app.animator.is_being_animated(switch, "_slider_position")


async def test_switch_does_not_animate_on_none() -> None:
    app = SwitchApp()
    app.animation_level = "none"

    async with app.run_test() as pilot:
        switch = app.query_one(Switch)
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        switch.action_toggle_switch()
        await pilot.pause()
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        # Move to the next frame.
        animator()
        # The animation should still be running.
        assert not app.animator.is_being_animated(switch, "_slider_position")



from textual.app import App, ComposeResult
from textual.widgets import Static


TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain."""


class BoxSizing(App):
    def compose(self) -> ComposeResult:
        self.widget1 = Static(TEXT)
        yield self.widget1
        self.widget2 = Static(TEXT)
        yield self.widget2

    def on_mount(self) -> None:
        self.widget1.styles.background = "purple"
        self.widget2.styles.background = "darkgreen"
        self.widget1.styles.width = 30
        self.widget2.styles.width = 30
        self.widget1.styles.height = 6
        self.widget2.styles.height = 6
        self.widget1.styles.border = ("heavy", "white")
        self.widget2.styles.border = ("heavy", "white")
        self.widget1.styles.padding = 1
        self.widget2.styles.padding = 1
        self.widget2.styles.box_sizing = "content-box"


if __name__ == "__main__":
    app = BoxSizing()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import VerticalScroll, Container
from textual.widgets import Button


class ScrollVisibleMargin(App):
    CSS = """
    Container {
        height: auto;
        margin-top: 8;
        border: solid red;
    }
    """

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            with Container():
                for index in range(1, 51):
                    yield Button(f"Hello, world! ({index})", id=f"b{index}")

    def key_x(self):
        button_twenty = self.query_one("#b26")
        button_twenty.scroll_visible()


app = ScrollVisibleMargin()
if __name__ == "__main__":
    app.run()



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import TabbedContent, DataTable


class Dashboard(App):
    """Dashboard"""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        Binding("ctrl+q", "app.quit", "Quit", show=True),
    ]
    TITLE = "Dashboard"
    CSS = """

    DataTable {
        height: auto;
    }
    """

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        with TabbedContent("Workflows"):
            yield DataTable(id="table")

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_columns("Id", "Description", "Status", "Result Id")
        for row in [(1, 2, 3, 4), ("a", "b", "c", "d"), ("fee", "fy", "fo", "fum")]:
            table.add_row(key=row[0], *row)


if __name__ == "__main__":
    app = Dashboard()
    app.run()
